#!/usr/bin/env python

# Copyright (C) 2002-2006
# Max-Planck-Institut fuer Radioastronomie Bonn
#
# Produced for the ALMA and APEX projects
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Library General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option) any
# later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Library General Public License for more
# details.
#
# You should have received a copy of the GNU Library General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 675 Massachusetts Ave, Cambridge, MA 02139, USA.  Correspondence concerning
# ALMA should be addressed as follows:
#
# Internet email: alma-sw-admin@nrao.edu
#
# Correspondence concerning APEX should be addressed as follows:
#
# Internet email: dmuders@mpifr-bonn.mpg.de
#
# "@(#) $Id: mbfits.py,v 1.342 2006/05/02 19:11:23 vltsccm Exp $"
#
# The Python ALMA-TI FITS / APEX MBFITS Raw Data Writer
#
# Who			When		What
#
# D. Muders, MPIfR	04/2002		Initial version
# H. Beuther, MPIfR	06/2002		Additions for benchmarking
# D. Muders, MPIfR	08/2002		Changes for arrays and strings
#					after pCFITSIO fix
# D. Muders, MPIfR	08/26/2002	Introduced classes for the FITS
#					file and for the SCAN-MBFITS
#					table. The goal is to add all
#					ALMA-TI and MBFITS tables.
# D. Muders, MPIfR	09/02/2002	Replaced fits_write_key with
#					fits_update_key as recommended
#					in CFITSIO.
#
#					Added addKeyword and addTableEntry
#					methods for SCAN-MBFITS table
#					class.
# C. Koenig, MPIfR      11/13/2002      Restructured object hirarchy
#                                       - XML functionality and new methods
#                                       were added
#                                       - creation of whole MBFits-Scans is
#                                       now possible including hirachy checking
#
# C. Koenig, D.Paurat, MPIfR   01/2003	Initial version of fits reader
#
# F. Schuller, MPIfR	01/28/2003	Read MBFITS file using
#					fitsio high level procedures
# F. Schuller, MPIfR	01/30/2003	Merge reader and writer modules
# F. Schuller, MPIfR	02/03/2003	Rewrite read methods without using XML objects
#					Define __str__ method for Header objects
# F. Schuller, MPIfR	02/04/2003	Add __str__ methods for BinTable, Table and FitsFile
# F. Schuller, MPIfR	02/10/2003	Move subclasses to 'normal' classes

# F. Schuller, MPIfR    02/14/2003      Method BinTable.readTableColumn implemented
# F. Schuller, MPIfR	03/05/2003	Primary Header keywords moved to XML description
# F. Schuller, MPIfR	03/13/2003	Handle two-dimensional data in BinTable.readTableColumn
# C. Koenig, MPIfR      03/20/2003      Moved XmlStructure to a seperate moudule
#                                       Added getObservation method
# F. Wyrowski, MPIfR    05/23/2003      Changed OBSMODE into SCANTYPE
# C.Koenig,MPIfR	06/03/2003	Create method of TblElement and BinTable have
#					been modified for correct datatype handling
# F. Wyrowski, MPIfR    06/11/2003      getObsmode added. modification of readTableColumn
#                                       to correctly read multiple entries in BinTables
# F. Wyrowski, MPIfR    06/11/2003      modification of readTableColumn to get
#                                       rid of list of list of strings
# F. Schuller, MPIfR    06/11/2003      one more correction in readTableColumn
#                                       to read 2D data
# F. Schuller, MPIfR    06/17/2003      - corrections in datatypes when reading
# + E. Polehampton                      - remove 2nd dimension in 1D read arrays
# F. Schuller, MPIfR    06/25/2003	reshape data in readTableColumn when multidimensional
#					and remove the TDIM field in BinTable.create when only 1D
# F. Schuller, MPIfR    06/27/2003      use the config.DEBUG global variable to control the
#                                       amount of debugging infos to be displayed
# F. Schuller, MPIfR    06/30/2003      method BinTable.readTableRow implemented
# F. Schuller, MPIfR	07/01/2003	added a test before reshaping data in readTableColumn
# C.Koenig, MPIfR	07/07/2003	method FitsFile.getTables implemented
# C.Koenig, MPIfR       07/17/2003      FitsFile.__init__ now allowes usage of named XML description
# C.Koenig, MPIfR       08/06/2003      Changed "python2.2" to "python"
# C.Koenig, MPIfR       08/08/2003      Default XML file will now be searched in the MBFITSPATH
#					(environment variable of os)
# C.Koenig, MPIfR       08/11/2003      argument XmlFileName has been replaced by xmlType
#					(as defined in XmlStructure.py)
# F. Schuller, MPIfR	08/13/2003	changed the getObservation method to return
#                                       a list of Hdu numbers and table names
# C.Koenig, MPIfR       08/15/2003      - added method readKeyword to class Header
#                                       - added method lookForChanges to class TblElement
#                                       - changed __init__ methods of classes BinTable, Header,
#					TblElement so they require a parent argument - parent is
#					stored in self.parent
#                                       - changed method create of class Header so it now takes
#					care of the changeLookUp XML attribute
# F. Schuller, MPIfR	08/21/2003	removed the fptr argument from all methods: it's now
#					an attribute of each Table, Header and BinTable object
# C.Koenig, MPIfR       08/27/2003      - correction to FitsFile.getTables method
#                                       - correction to BinTable.addTableRow for writing multidimensional entries
# C.Koenig, MPIfR       09/02/2003      import of old apex_connector module is now replaced by import of
#                                       the connectorModule named in the XML file - global connectorModule
#                                       is now called 'connector' (instead of apex_connector)
# E.Polehampton         11/21/2003      Update getObservation() with the possibility to specify the BaseBand.
#                                       Put some print statments in read() into DEBUG mode.
# F. Schuller, MPIfR    12/03/2003      corrections in TblElement.__init__ and Header.create
#                                       for proper datatype initialisation
# F. Schuller, MPIfR    12/09/2003      changed the description of 'Primary' in the xml file
#                                       to a header within a table, so that the updateKeyword
#                                       method of our Header class can be used
# F. Schuller, MPIfR    12/11/2003      differentiate 'E' and 'D' types in BinTable.readTableColumn
# C.Koenig, MPIfR       02/14/2004      - Transition to use cfitsio instead of pcfitsio
#                                       - Added functions for file pointer handling
# C.Koenig, MPIfR       02/11/2004      - Removed open() / close() calls from FitsFile.read() and FitsFile.getObservation()
#                                         FitsFiles should always be used the following way:
#
#                                         myFile = FitsFile('someFitsFile.fits')
#                                         myFile.open()
#                                         ...
#                                         <operate on the file>
#                                         ...
#                                         myFile.close()
#
#                                         You should not try to close it when you are about to use file accessing methods of
#                                         some subclases (like Table.BinTable.readTableRow()). Although this is possible and each
#                                         file accessing method checks for a valid filepointer when it is called, it might cause
#                                         segmentation faults when calling the cfitsio routines, esp. in multithreaded environment
#                                         environments, when the file might be closed during the execution of a file accessing method.
# F. Schuller, MPIfR    03/19/2004      Changed the Header.addKeyword method to be able to add keywords not defined in
#                                       the MB-Fits definition (temporary solution while using MB-Fits ver. 1.52, the
#                                       missing keywords should exist in the next version)
# F. Schuller,   MPIfR  05/27/2004      changed the method readTableColumn - much faster
# F. Schuller,   MPIfR  05/29/2004      some bug fixed in BinTable methods
# E.Polehampton, MPIfR  06/03/2004      changed to numarray
# E.Polehampton, MPIfR  06/09/2004      add method to FitsFile: makeHeaderKeyList() - this used to be makeObsList()
#                                         in CalibController.
#                                         change getObsmode() in FitsFile to actually 'get' the obsMode.
# E.Polehampton, MPIfR  06/24/2004      fix in readTableRow (numarray) and readTableColumn - need to account for
#                                       variable length array that may have length 1
# C.Koenig,      MPIfR  06/29/2004      fix in Header.create() - type of keywords was not written the right way
# C.Koenig,      MPIfR  07/01/2004      - fix in TblElement.lookForChanges() - elements were identified as being
#                                         changed although nothing has changed
#                                       - added method setVariableLengthDimension(colName,length) to BinTable
# C.Koenig,      MPIfR  07/05/2004      added lookForChanges calls into methods Header.addKeyword and Header.updateKeyword
# C.Koenig,      MPIfR  08/03/2004      fix in TblElement.lookForChanges()
# C.Koenig,      MPIfR  17/01/2005      fix in BinTable.readTableColumn()
# E.Polehampton  MPIfR  11/03/2005      fix in updateDimension()
# C.Koenig,      MPIfR  14/03/2005      _getColDimensions: removed the calculation for the " / " as
#                                       string array field seperator
# E.Polehampton  MPIfR  17/03/2005      - update setVariableLengthDimension, updateDimension() and _getColDimensions()
#                                          to allow for multi-dimension variable-length-arrays and to properly
#                                          dimension columns with strings.
#                                       - Comment out ComClass
# D.Muders       MPIfR  19/04/2005      Trimmed off white space of variable length dimension string
# F. Schuller    MPIfR  13/06/2005      - Minor changes (e.g. converting strings to int, removing
#                                         extra quotes) to make it compatible with Python 2.3 on Mac OS X
#                                       - comment out references to connector
# E.Polehampton  MPIfR  15/06/2005      Finally comment out all "dataSource", ComClass, connector stuff.
#
# R.Schaaf       MPIfR  11/07/2005      - Added FitsFile.writeChecksums that
#                                       writes (or updates) the checksums of
#                                       all HDUs
# E.Polehampton  MPIfR  11/07/2005      - change addKeyword so that it can add keywords that are not
#                                         in the xml description
#                                       - corrections to some error messages
# R.Schaaf       MPIfR  15/07/2005      global variable XmlStruc in class
#                                       FitsFile caused memory leak, because
#                                       it was newly created for each new
#                                       FitsFile object. Fixed memory leak
#                                       by introducing global dictionary
#                                       XmlStructDict.
# R.Schaaf       MPIfR  01/09/2005      Changed BinTable.addTableRow and
#                                       BinTable.addTableRow to accept
#                                       Numeric.arrays in addition to lists.
# E.Polehampton  MPIfR  18/09/2005      change readTableColumn, readTableRow to return Numeric
#                                       array not numarray.
# R.Schaaf       MPIfR  04/11/2005      Header.unpdateKeyword: Truncate string
#                                       values that are too long
# R.Schaaf       MPIfR  18/11/2005      BinTable.addTableRows: Modified to
#                                       accept Numeric.arrays; still not
#                                       complete
# R.Schaaf       MPIfR  06/12/2005      Marked the methods BinTable._getTTYPE,
#                                       _getTFORM, _getTDIM, _getDataDimension,
#                                       and updateDimension as broken
# R.Schaaf       MPIfR  06/12/2005      Added specialized methods
#                                       BinTable.addTableRowsDatapar and
#                                       .addTableRowsArraydata for performant
#                                       writing to these tables.
# R.Schaaf       MPIfR  08/12/2005      Added method
#                                       BinaryTable.getOptNumRowsToAdd
# D.Muders       MPIfR  09/01/2006      - Removed initial CREATOR keyword generation.
#					- Write optional file generation comment.
# R.Schaaf       MPIfR  03/24/2006      General structure of FitsFile changed.
#                                       A FitsFile object is not longer related
#                                       to a single (physical) fits file, but
#                                       instead to several physical fits files,
#                                       each one containing one or more
#                                       (physical) HDUs. The fits file containing
#                                       the FitsFile object's Primary Table may
#                                       contain one or more Grouping tables, which
#                                       hold links to the various HDUs according to
#                                       FITSIO's Hierarchical Grouping mechanism
#                                       (See "CFITSIO User's Refernce Guide"
#                                       Chapter 8 and references therein.)
#
#                                       The interface of mbfits was kept
#                                       as unchanged as possible: Table.create
#                                       has a new (optional) argument that specifies
#                                       the file in which the Table resides. Also,
#                                       the interpretaion of argument "filename" in
#                                       FitsFile.__init__ was changed: If filename
#                                       specifies a directory, the new structure
#                                       is assumed. Otherwise, the old structure is
#                                       used. This allows FitsFiles created with
#                                       the old singlefile structure to be read
#                                       by mbfits. However, FitsFiles are allways
#                                       written in the new multifile structure.
#
#                                       The interface of the module is not clean,
#                                       since it exposes implementation details
#                                       that do no longer have their original
#                                       meaning: FitsFile.fptr is exposed by
#                                       FitsFile.create, .open, and .fileHandle. In
#                                       a multifile structure it can not longer
#                                       have the meaning it had in the singlefile
#                                       structure. Also, Table.HduNum is exposed
#                                       by various methods. Again, the meaning of
#                                       this variable can not longer be the same.
#
#                                       The way out of this dilemma is the directory
#                                       FitsFile.HduMap, which relates the
#                                       (logical) HDU number Table.HduNum with
#                                       a file pointer and a (physical) HDU number.
#                                       These quantities give the correct HDU.
#                                       Clients, that use file pointers and HDU
#                                       numbers to access tables directly, must
#                                       use FitsFile.HduMap to retrieve the
#                                       correct file pointer and HDU number before
#                                       calling cfitsio routines!
#
# R.Schaaf       MPIfR  03/27/2006      Added method Table.getFilename
#
# R.Schaaf       MPIfR  03/29/2006      . _HierGroup: Some workarounds for
#                                       broken routines in p_cfitsio removed
#                                       after fixing p_cfitsio.i.
#                                       . _HierGroup: Columns with
#                                       indexing information (EXTNAME, SUBSNUM,
#                                       FEB, BASEBAND) added.
#                                       FitsFile.addTable and Table.create
#                                       modified to pass this indexing
#                                       information.
# R.Schaaf       MPIfR  03/30/2006      Added method
#                                       mergeMBGroupFitsToSingleFile that
#                                       copies a multifile structured
#                                       Fitsfile into a singlefile structured
#                                       one.
# H.Hafok       , MPIfR 10/04/2006      changed order in ARRAY_DATA, shape command was wrong.
# C.Koenig,      MPIfR  05/02/2006      - changed looping structure in Header.updateKeyword
#                                         for better performance
#                                       - fixed loop exit condition in Header.updateKeyword


# ----------- TO DO -----------
#
# 
#


""" This is the ALMA-TI FITS / MBFITS Files Tools Library.
(c) 2002-2003 MPIfR """

import cfitsio
#import numarray
import Numeric
import string
from XmlStructure import *
import config
import os

#<timing test>
import time
#</timing test>

# Initialize some global variables
# connector = None
# typeDict = {'L':14,'A':16,'I':21,'J':41,'E':42,'D':82}
XmlStructDict = {}



#########################################################################
class FitsFile:
#########################################################################

    """ The main MBFITS file class.\n"""

    def __init__(self,filename, xmlType='MBFITS'):

        """
        Initialize a MBFITS File object

        Parameters:
        filename: name of (logical) fits file to be created/opened. If filename
                  specifies a directory, a multifile structure in this directory
                  is assumed. The physical fits file filename+'/GROUPING.fits' is
                  related with the Primary Table. Otherwise, the original singlefile
                  structure in filename is assumed. 
        xmlType: type for the XML description to be used as defined in
                 XmlStructure.py (default type is 'MBFITS')
        """

        global XmlStruc
        if not xmlType in XmlStructDict:
            XmlStructDict[xmlType] = XmlStructure(xmlType)
        XmlStruc = XmlStructDict[xmlType]

	self.currentFileName = filename
        self.fptr = None
        self.HduCount = 0
	self.TableList = []
        self.ChildTables=[] 	# list of tables related to the FitsFile object
	self.Name = 'FitsFile'	# this is the "TableName" of a FitsFile object, should
              # be changed to FitsFile, but needs to be changed in the XML file as well
              # (stg like: SCAN-MBFITS needs a Parent of type FitsFile)

        self.FitsFile = self

        # Variables related with the multifile structure:
        #
        # The physical filename
        if os.path.isdir(self.currentFileName):
            # multifile structure:
            self._pFilename = self.currentFileName + "/GROUPING.fits"
        else:
            # singlefile structure:
            self._pFilename = self.currentFileName

        # The HduMap:
        # (physical file pointer, physical HDU number)[logical HDU number] 
        self.HduMap = {}


    def __str__(self):
	if len(self.TableList) == 0:
	    result = "No Table found in this file"
	else:
	    result = str(len(self.TableList))+" Tables found...\n"
	    for tab in self.TableList:
		result = result+tab.__str__()
	return result



    def create(self, comment = ''):

        """ Creates an MBFITS Raw Data file
	initializing it with the default primary image HDU
	and MBFITS keywords. Returns a pointer to the file. """

        # Create a new FITS file
        if config.DEBUG:
            print 'Creating file'

        status,fptr = cfitsio.fits_create_file(self._pFilename)
        if status:
            msg = 'Could not create file', self._pFilename
            myError = MBFITSError(msg,code=status)
            raise myError
            
        self.fptr = fptr

        # The FITS standard prescribes an image as first HDU.
	# Create an empty one and add MBFITS keywords.
        status=cfitsio.fits_create_img(fptr,32,0,[])

	# Use the XML description to generate the Primary Header
        table=XmlStruc.XmlScan.tables['Primary']
        tblHeader=table.getElementsByTagName('HEADER')[0]
	# primary = Header(self)
        self.HduCount = 1
        self.HduMap[1] = (fptr, 1)
        
	primary = Table("Primary",self)
        primary.HduNum = 1
	primary.Header.create(tblHeader.childNodes)

	# Add optional comment (e.g. for files taken with the simulator)
	if (comment != ''):
            status=cfitsio.fits_write_comment(fptr, comment)

        self.ChildTables.append(primary)
        self.TableList.append(primary)

        # Create the grouping table:
        hg = _HierGroup()
        hg.createGroup(fptr, "GROUP_TABLE")

#<debug>
##         status, nhdus = cfitsio.fits_get_num_hdus(self.fptr)
##         print "RS_DEBUG FitsFile.create nhdus:", nhdus
#</debug>

        self._updateHduMaps(self.HduMap)

        return fptr



    def open(self,mode=0):

        """ Open an APEX MBFITS Raw Data file. """

        # Open a FITS file in mode 0 (read-only) or 1 (read-write)

 	filename = self._pFilename
        status, fptr = cfitsio.fits_open_file(filename, mode)
        if status == 0:
            # Get number of physical HDUs in this physical file:
            status, pHduCount = cfitsio.fits_get_num_hdus(fptr)
            if status:
                msg = "Error while opening file %s"%(filename)
                myError = MBFITSError(msg)
                raise myError
                
            # Walk through the file to determine logical HduNumbers,
            # build the HduMap, and open physical files that contain members of
            # groups:
            lHduNum = 1
            for pHduNum in xrange(1, pHduCount+1):

#<debug>
##                 print "RS_DEBUG FitsFile.open lHduNum, pHduNum:", lHduNum, pHduNum 
#</debug>

                status, cpHduNum = cfitsio.fits_movabs_hdu(fptr, pHduNum)
                if status:
                    msg = "Error while opening file %s"%(filename)
                    myError = MBFITSError(msg)
                    raise myError

                hg = _HierGroup()
                if hg.isGroupingTable(fptr):
#<debug>
##                     print "RS_DEBUG Grouping table!"
#</debug>
                    groupHduList = hg.openAllMembers(fptr, filename, mode)
#<debug>
##                     print "RS_DEBUG groupHduList:", groupHduList
#</debug>
                    for (fptrMember, pHduNumMember) in groupHduList:
#<debug>
##                         print "RS_DEBUG lHduNum, pHduNumMember:", lHduNum, pHduNumMember
#</debug>
                        self.HduMap[lHduNum] = (fptrMember, pHduNumMember)
                        lHduNum += 1
                else:
                    self.HduMap[lHduNum] = (fptr, pHduNum)
                    lHduNum += 1
                    
            self.fptr = fptr
            self.HduCount = lHduNum-1

#<debug>
##             print "RS_DEBUG self.HduCount, self.HduMap:", self.HduCount, self.HduMap
#</debug>

            # Update all HduMaps
            self._updateHduMaps(self.HduMap)

	    # Determine file and XML MBFITS versions.
            fptr, pHduNum = self.HduMap[1]
            status = cfitsio.fits_movabs_hdu(fptr,pHduNum)

	    FileMBFITSVersion = \
                cfitsio.fits_read_keyword(fptr, 'MBFTSVER')[1]
	    FileMBFITSVersion = FileMBFITSVersion[1:-2].rstrip()

            primaryHDU = XmlStruc.XmlScan.tables['Primary']
            primaryHDUHeader = primaryHDU.getElementsByTagName('HEADER')[0]
            primaryHDUHeaderStringKeywords = primaryHDUHeader.getElementsByTagName('A')

            XmlMBFITSVersion = ''

            for keyword in primaryHDUHeaderStringKeywords:
                if (keyword.childNodes[0].nodeValue == 'MBFTSVER'):
                    XmlMBFITSVersion = keyword.getAttribute('preset')
                    break

            if (FileMBFITSVersion != XmlMBFITSVersion):
		print 'WARNING: MBFITS file version (%s) is differs from XML version (%s).' % (FileMBFITSVersion, XmlMBFITSVersion)
        else:
            msg = "Could not open file %s"%(filename)
            myError = MBFITSError(msg)
            raise myError

        return fptr



    def close(self):

        """ Close current APEX MBFITS Raw Data file. """

	# This method should also delete (from memory) all the Table, Header
	# and BinTable objects in the hierarchy under 'self'

        closedFptrs = []

        for lHduNum in self.HduMap.keys():
            fptr, pHduNum = self.HduMap[lHduNum]
            if not fptr in closedFptrs: 
                closedFptrs.append(fptr)
                status = cfitsio.fits_close_file(fptr)
                if status:
                    msg = "Could not close file for HduNum %d - status %d" \
                          % (lHduNum, status)
                    myError = MBFITSError(msg)
                    raise myError
                
        if not self.fptr in closedFptrs: 
            closedFptrs.append(self.fptr)
            status = cfitsio.fits_close_file(self.fptr)
            if status:
                msg = "Could not close FitsFile %s - status %d" \
                      % (self._pFilename, status)
                myError = MBFITSError(msg)
                raise myError

        self.fptr = None
        self.HduMap = None

        self._updateHduMaps(self.HduMap)


    def addTable(self, table, fptrTable, pHduNumTable, **kargs):
        """
        Adds a Table object to the grouping table.

        **args is indexing information passed to _HierGroup.addMember
        """

        # Move to grouping table:
        status, hdu2 = cfitsio.fits_movabs_hdu(self.fptr,2)
        if status:
            msg = "Could not move to grouping table - status", status
            myError = MBFITSError(msg)
            raise myError

        hg = _HierGroup()
        hg.addMember(self.fptr, fptrTable, pHduNumTable, **kargs)

        lHduNumTable = self.HduCount + 1 
        self.HduMap[lHduNumTable] = (fptrTable, pHduNumTable)
        self.TableList.append(table)
        self.HduCount = lHduNumTable
#<debug>
##         print "RS_DEBUG FitsFile.addTable table.Name, pHduNumTable, lHduNumTable, self.HduCount, self.HduMap:", table.Name, pHduNumTable, lHduNumTable, self.HduCount, self.HduMap
#</debug>

        return lHduNumTable
        
    def fileHandle(self):

        """ Return CFITSIO file handle. """

        return self.fptr



    def _updateHduMaps(self, hduMap):
        
        """ Update file pointers in all associated tables """
        
        for table in self.ChildTables:
            table._updateHduMaps(hduMap)



    def read(self):

	""" Read an MBFITS file and store all the data
        in the TableList attribute """
        
	if not self.fptr:
            msg = 'No file pointer. File seems to be closed.'
            myError = MBFITSError(msg)
            raise myError

        HduCount = self.HduCount
        fptr = self.fptr
	# First, read the Primary Header
        if config.DEBUG:
            print "Now reading Primary Header"
	Primary = Table('Primary',self)
        Primary.HduNum=1
	Primary.Header.read()
	self.TableList.append(Primary)

	# Then read all the tables
	for HduNum in xrange(2,HduCount+1):
	    table = Table(Parent=self)
	    table.read(HduNum)
	    self.TableList.append(table)
        


    def makeHeaderKeyList(self, tablename, keyword, FeBe='*',BaseBand='*',uniq=0):
        """
        Examines the FITS file header to make a list of values of
	a given keyword in the Scan. You can specify the FEBE and
        BaseBand to restrict the search to one of these.
        uniq is a logical that determines whether the final list contains ALL values
        of the keyword (=0), or just the unique ones (=1)

        If the FitsFile has not previously been read then only Headers are read.

        input:
               tablename  = string: Name of Table (e.g. 'ARRAYDATA-MBFITS')
               keyword    = string: Header keyword to match (e.g. 'OBSNUM')
               FeBe       = string: FeBe name (optional)
               BaseBand   = integer: BaseBand Number (optional)
               uniq       = logical (0 or 1)
        output:
               list  = list of values of keyword
        """
        # if the file has not already been read then read all the Headers
        if len(self.TableList) == 0:
            if not self.fptr:
                msg = 'No file pointer. File seems to be closed.'
                myError = MBFITSError(msg)
                raise myError
        
        
            HduCount = self.HduCount

            # First, read the Primary Header            
            Primary = Table('Primary',self)
            fptr, pHduNum = self.HduMap[1]
            status = cfitsio.fits_movabs_hdu(fptr,pHduNum)
#<test>
            Primary.HduNum = 1
#</test>
            Primary.Header.read()
            self.TableList.append(Primary)
        
            # Then read the Headers of other tables
            for HduNum in xrange(2,HduCount+1):
                fptr, pHduNum = self.HduMap[HduNum]
                status = cfitsio.fits_movabs_hdu(fptr,pHduNum)
                # read table name
                status,Name,comment = cfitsio.fits_read_keyword(fptr,'EXTNAME')
                # Remove the leading and trailing quotes
                Name = Name[1:len(Name)-1]
                table = Table(Name,Parent=self)
#<test>
                table.HduNum = HduNum
#</test>
                table.Header.read()
                self.TableList.append(table)                    

        # Now, try and find the require keyword and value to make the list
        # loop over tables
        
        KeyValList = []
        for table in self.TableList:
            GoAhead = 0
            if table.Name == tablename and table.Header.KeywordDict.has_key(keyword):
                # if the table to be searched is ARRAYDATA or DATAPAR then check for
                # FeBe (and BaseBand)
                if table.Name == 'DATAPAR-MBFITS':
                    if FeBe == '*' or table.Header.KeywordDict['FEBE'][0] == FeBe:
                        GoAhead = 1
                elif table.Name == 'ARRAYDATA-MBFITS':                    
                    if FeBe == '*' or table.Header.KeywordDict['FEBE'][0] == FeBe:
                        if BaseBand == '*' or table.Header.KeywordDict['BASEBAND'][0] == BaseBand:
                            GoAhead = 1
                else:
                    GoAhead = 1

                if GoAhead:
                    keyval = table.Header.KeywordDict[keyword]
                    # if uniq is 1 then only key values that are different will be appended
                    if uniq == 1:
                        if keyval[0] not in KeyValList:
                            KeyValList.append(keyval[0])
                    else:
                        KeyValList.append(keyval[0])
            
        print keyword+" found: "+str(KeyValList)
        return KeyValList
        


    def showObsmode(self):
	"""
        Print the Observing mode(s) found in the SCAN-MBFITS table(s)
        This requires that the FitsFile has been read().
        """

	for Table in self.TableList:
	    if Table.Name == "SCAN-MBFITS":
		Table.showObsmode()



    def getObsmode(self):
 
        """
        getObsmode() --> ObsMode    (string)
        
        Get the Observing mode(s) found in the SCAN-MBFITS table(s)
        This comes from the SCANTYPE keyword in MBFITS.
        This uses cfitsio to actually GET the keyword.
        It allows you to get the ObsMode without reading the entire file.
        If the file has already been read then just show the result by using
        showObsmode().
        If no SCAN-MBFITS table is found the value of Obsmode will be None.
        """

        if not self.fptr:
            msg = 'No file pointer. File seems to be closed.'
            myError = MBFITSError(msg)
            raise myError
        
        HduCount = self.HduCount

        ObsMode = None   # initialisation, will contain SCANTYPE
        
        for HduNum in xrange(2,HduCount+1):
            fptr, pHduNum = self.HduMap[HduNum]
	
            # Move to table
            status = cfitsio.fits_movabs_hdu(fptr,pHduNum)

            # read table name
            status,Name,comment = cfitsio.fits_read_keyword(fptr,'EXTNAME')
            # Remove the leading and trailing quotes
            Name = Name[1:len(Name)-1]
  
            # when find Scan table, get the Obsmode and return it.
            if Name == "SCAN-MBFITS":
                status,ObsMode,comment = cfitsio.fits_read_keyword(fptr,'SCANTYPE')
                # Remove the leading and trailing quotes and spaces
                ObsMode = string.rstrip(ObsMode[1:len(ObsMode)-1])

                return ObsMode

        # if Scan table not found then return None
        ObsMode = None
        return ObsMode


    def getInstrument(self):
 
        """
        getInstrument() --> Instrument   (string)
        
        Get the instrument from the Primary header - should be
        one of APEXHET or APEXBOL.
        If not found, the returned value is None.
        """

        if not self.fptr:
            msg = 'No file pointer. File seems to be closed.'
            myError = MBFITSError(msg)
            raise myError
        
        Instrument = None   # initialisation, will contain INSTRUME
        
        # Move to Primary header
        fptr, pHduNum = self.HduMap[1]
        status = cfitsio.fits_movabs_hdu(fptr,pHduNum)
        try:
            status,instrum,comment = cfitsio.fits_read_keyword(fptr,'INSTRUME')
            # Remove the leading and trailing quotes and spaces
            Instrument = string.rstrip(instrum[1:-1])
        except KeyError:
            pass  # result already contains None

        return Instrument


    def getObservation(self,ScanNum,ObsNum,FeBe,BaseBand=1):

        """ Find tables corresponding to one observation represented by
        ScanNum, ObsNum and FeBe
        
        returns a list of tuples, where:
        Result[0] = ('Obsmode',obs. mode)
        Result[i>0] = (Hdu number, table name)"""

        if not self.fptr:
            msg = 'No file pointer. File seems to be closed.'
            myError = MBFITSError(msg)
            raise myError

        HduCount = self.HduCount

        TableList = []   # will contain the (HduNum,Name) tuples
        ObsMode = None   # initialisation, will contain SCANTYPE
        
        # Get all Tables with same ScanNum, ObsNum and FeBe combination
        for HduNum in xrange(2,HduCount+1):
	    SCANNUM = None
            OBSNUM = None
            FEBE = None
            BASEBAND = None

            fptr, pHduNum = self.HduMap[HduNum]

            # Move to table
            status = cfitsio.fits_movabs_hdu(fptr,pHduNum)

            # read table name
            status,Name,comment = cfitsio.fits_read_keyword(fptr,'EXTNAME')
            # Remove the leading and trailing quotes
            Name = Name[1:-1]

            # read SCAN-NUM
            status,SCANNUM,comment = cfitsio.fits_read_keyword(fptr,'SCANNUM')
            SCANNUM = _setType(SCANNUM)
            SCANNUM = int(SCANNUM)

            # Try to read OBS-NUM
            try:
                status,OBSNUM,comment = cfitsio.fits_read_keyword(fptr,'OBSNUM')
                OBSNUM = _setType(OBSNUM)
                if OBSNUM == "":
                    OBSNUM = ObsNum
                OBSNUM = int(OBSNUM)
            except:
                OBSNUM = ObsNum
            # Try to read FEBE Combination
            try:
                status,FEBE,comment = cfitsio.fits_read_keyword(fptr,'FEBE')
                FEBE = _setType(FEBE)
                if FEBE == "":
                    FEBE = FeBe
                if FEBE[0] == "'":
                    FEBE = FEBE[1:-1]
            except:
                FEBE = FeBe
            # Try to read BaseBand - if not specified look for BaseBand=1
            try:
                status,BASEBAND,comment = cfitsio.fits_read_keyword(fptr,'BASEBAND')
                BASEBAND = _setType(BASEBAND)
                if BASEBAND == "":
                    BASEBAND = BaseBand
                BASEBAND = int(BASEBAND)
            except:
                BASEBAND = BaseBand
            if (SCANNUM == ScanNum) and (OBSNUM == ObsNum) and (FEBE == FeBe) and (BASEBAND == BaseBand):
                TableList.append((HduNum,Name))
                    
            # Get ObsMode for this Observation
            if Name == "SCAN-MBFITS":
                status,ObsMode,comment = cfitsio.fits_read_keyword(fptr,'SCANTYPE')
                # Remove the leading and trailing quotes and spaces
                ObsMode = string.rstrip(ObsMode[1:len(ObsMode)-1])
	
        # fill Result
        Result = [('ObsMode',ObsMode)]
        Result.extend(TableList)
        
        return Result



    def getObservationTables(self,ScanNum,ObsNum,FeBe,BeSec=1):

        """
        Get all tables for this observation.

        Usage: myFitsFile.getObservationTables(ScanNum <int>, ObsNum <int>, FeBE <str> [, backend section <int>])

        -----
        Returns an observation object, that has the tables as its properties:

        observation.scan:      the SCAN-MBFITS table
        observation.febe:      the FEBEPAR-MBFITS table
        observation.arraydata: the ARRAYDATA-MBFITS table
        observation.datapar:   the DATAPAR-MBFITS table
        observation.monitor:   the MONITOR-MBFITS table
        """

        #########################################################################
        class Observation:
        #########################################################################
            
            def __init__(self, file, scan, obs, febe, BeSec = 1):
                self.scan = None
                self.febe = None
                self.arraydata = None
                self.datapar = None
                self.monitor = None

                self.scannum = scan
                self.obsnum = obs
                self.febename = febe
                obs = file.getObservation(scan, obs, febe, BeSec)
                tables = obs[1:]
                for tabInfo in tables:
                    name = tabInfo[1]
                    hdu = tabInfo[0]

                    if name == 'SCAN-MBFITS':
                        Parent = file
                        self.scan = self.__createTable(name,Parent,hdu)
                    elif name =='FEBEPAR-MBFITS':
                        Parent = self.scan
                        self.febe = self.__createTable(name,Parent,hdu)
                    elif name =='ARRAYDATA-MBFITS':
                        Parent = self.febe
                        self.arraydata = self.__createTable(name,Parent,hdu)
                    elif name =='DATAPAR-MBFITS':
                        Parent = self.febe
                        self.datapar = self.__createTable(name,Parent,hdu)
                    elif name =='MONITOR-MBFITS':
                        Parent = self.scan
                        self.monitor = self.__createTable(name,Parent,hdu)

                    ##################### end of class Observation ##############################

            def __createTable(self,name,Parent,hdu):
                table = Table(name,Parent)
                table.HduNum = hdu
                return table

        obs = Observation(self,ScanNum,ObsNum,FeBe,BeSec)
        return obs



    def observationExists(self,ScanNum,ObsNum,FeBe):

        """ Returns 1 if observation exists, 0 if not."""
        
        if not self.fptr:
            msg = 'No file pointer. File seems to be closed.'
            myError = MBFITSError(msg)
            raise myError
        
        HduCount = self.HduCount

        # Search through all tables for same ScanNum, ObsNum and FeBe combination
        Result = 0
        for HduNum in xrange(2,HduCount+1):
	    SCANNUM = None
            OBSNUM = None
            FEBE = None

            fptr, pHduNum = self.HduMap[HduNum]
	
            # Move to table
            status = cfitsio.fits_movabs_hdu(fptr,pHduNum)

            # read SCAN-NUM
            status,SCANNUM,comment = cfitsio.fits_read_keyword(fptr,'SCANNUM')
            SCANNUM = _setType(SCANNUM)

            # Try to read OBS-NUM
            try:
                status,OBSNUM,comment = cfitsio.fits_read_keyword(fptr,'OBSNUM')
                OBSNUM = _setType(OBSNUM)
            except:
                OBSNUM = None
            # Try to read FEBE Combination
            try:
                status,FEBE,comment = cfitsio.fits_read_keyword(fptr,'FEBE')
                FEBE = _setType(FEBE)
            except:
                FEBE = None
            if (SCANNUM == ScanNum) and (OBSNUM == ObsNum) and (FEBE == FeBe):
                Result = 1
                break
        
        return Result
        


    def getTables(self, Name, **Keywords):
        
        """
        Returns a list of all tables with same Name and matching
        TableHeader keywords

        usage:
        MyFile.getTable(TableName, key=value)

        TableName: Name of table to be searched for
        key=value: Header keyword and value pairs, wich should be
        matched (can be multiple pairs of keyword=value)
        """

        def _checkValue(fptr, Keyword, Value):
            """ Checks if the value of a keyword matches the search"""
            try:
                status,TableValue,comment = cfitsio.fits_read_keyword(fptr,Keyword)
                TableValue = _setType(TableValue)
            except:
                TableValue = 0
            if TableValue == Value:
                return 1
            else:
                return 0

        if not self.fptr:
            msg = 'No file pointer. File seems to be closed.'
            myError = MBFITSError(msg)
            raise myError

        HduCount = self.HduCount
        TableList = []

#<debug>
##         print "RS_DEBUG FitsFile.getTables Name, Keywords:", Name, Keywords
##         print "RS_DEBUG FitsFile.getTables Name, HduCount:", Name, HduCount
#</debug>
        for HduNum in xrange(1,HduCount+1):
            fptr, pHduNum = self.HduMap[HduNum]

#<debug>
##             print "RS_DEBUG FitsFile.getTables HduNum, fptr, pHduNum:", HduNum, fptr, pHduNum
#</debug>

            matches=0
            status = cfitsio.fits_movabs_hdu(fptr,pHduNum)
            if Name != "" and (HduNum !=1):
                if _checkValue(fptr,'EXTNAME',Name):
                    matches=1
            if matches:
                for Key in Keywords:
                    if not _checkValue(fptr, Key, Keywords[Key]):
                        matches=0
                        break
                    else:
                        matches=1
            if matches:
                TableList.append(HduNum) # logical HduNum

#<debug>
##         print "RS_DEBUG FitsFile.getTables Name, TableList:", Name, TableList
#</debug>
        return TableList

                
    def writeChecksums(self):
        """
        Writes or updates (if necessary) the checksums of all HDUs
        """

        if not self.fptr:
            msg = 'No file pointer. File seems to be closed.'
            myError = MBFITSError(msg)
            raise myError

        HduCount = self.HduCount

        # Iterate over all logical HDUs and write the checksums:
        for HduNum in xrange(1,HduCount+1):
            fptr, pHduNum = self.HduMap[HduNum]

            status = cfitsio.fits_movabs_hdu(fptr,pHduNum)
            status = cfitsio.fits_write_chksum(fptr)



#########################################################################
class _HierGroup:
#########################################################################

    """
    Encapsulates the structure of a Grouping table.

    A _HierGroup object does not represent a single Grouping table, but
    instead the general structure of any Grouping table as used in MBFITS.
    """

    def __init__(self):
        """
        Set up structural information:
        """
        # Grouptype parameters from fitsio.h:
        #define GT_ID_ALL_URI  0
        #define GT_ID_REF      1
        #define GT_ID_POS      2
        #define GT_ID_ALL      3
        #define GT_ID_REF_URI 11
        #define GT_ID_POS_URI 12
        self._groupType = 12

        # Aditional index columns:
        self._indexColTtypes = ["EXTNAME", \
                                "SUBSNUM", \
                                "FEBE", \
                                "BASEBAND"]
        self._indexColTforms = ["30A", \
                                "1J", \
                                "30A", \
                                "1J"]
        self._indexColWriteFunctions = {
            "EXTNAME": cfitsio.fits_write_col_str,
            "SUBSNUM": cfitsio.fits_write_col_lng,
            "FEBE": cfitsio.fits_write_col_str, 
            "BASEBAND": cfitsio.fits_write_col_lng}
        self._indexColBlankValues = {
            "EXTNAME": "", \
            "SUBSNUM": -999,
            "FEBE": "", \
            "BASEBAND": -999}
    
    def isGroupingTable(self, fptr):
        """
        Checks the keyword EXTNAME.

        The routine operates on the current HDU in fptr.
        """

        status, extname, comment = \
                cfitsio.fits_read_keyword(fptr, 'EXTNAME')
        if status:
            return 0
        return extname=="'GROUPING'"

    def createGroup(self, fptr, groupName):
        """
        Create a grouping table at the end of fptr.
        """
        # Standard grouping table columns:
        status = cfitsio.fits_create_group(fptr, groupName, self._groupType)
        if status:
            msg = "Error %d while creating grouping table %s" \
                  % (status, groupName)
            myError = MBFITSError(msg)
            raise myError

        status, nCols = cfitsio.fits_get_num_cols(fptr)
        if status:
            msg = "Error %d while creating grouping table %s" \
                  % (status, groupName)
            myError = MBFITSError(msg)
            raise myError

        # Additional indexing columns:
        status = cfitsio.fits_insert_cols(fptr, \
                                          nCols+1, \
                                          len(self._indexColTtypes), \
                                          self._indexColTtypes, \
                                          self._indexColTforms)
        if status:
            msg = "Error %d while creating grouping table %s" \
                  % (status, groupName)
            myError = MBFITSError(msg)
            raise myError

    def addMember(self, fptr, fptrMember, hduNumMember, **kargs):
        """
        Adds a member specified by fptrMember and hduNumMember to the
        group in the current HDU in fptr.

        **kargs holds a dictionary that is used to fill the indexing columns.
        """

        status = cfitsio.fits_add_group_member(fptr, fptrMember, hduNumMember)
        if status:
            msg = "Error %d while adding member to grouping table" % status
            myError = MBFITSError(msg)
            raise myError

        status, numrows = cfitsio.fits_get_num_rows(fptr)
        if status:
            msg = "Error %d while adding member to grouping table" \
                  % status
            myError = MBFITSError(msg)
            raise myError

        # Fill indexing columns:
        for colName in kargs.keys():
            colValue = kargs[colName]
            if colValue==None:
                colValue = self._indexColBlankValues[colName]

            # String matching parameters from fitsio.h:
            #define CASESEN   1   /* do case-sensitive string match */
            #define CASEINSEN 0   /* do case-insensitive string match */
            status, colnum = cfitsio.fits_get_colnum(fptr, 0, colName)
            if status:
                msg = "Error %d while adding member to grouping table" \
                      % status
                myError = MBFITSError(msg)
                raise myError

            status = \
                self._indexColWriteFunctions[colName](fptr, \
                                                      colnum, \
                                                      numrows, \
                                                      1, \
                                                      1, \
                                                      [colValue])
            if status:
                msg = "Error %d while adding member to grouping table" \
                      % status
                myError = MBFITSError(msg)
                raise myError

#<debug>
##         status, nGroupMembers = \
##                 cfitsio.fits_get_num_rows(fptr)
##         print "RS_DEBUG _HierGroup.addMember fptr, fptrMember, hduNumMember, nGroupMembers:", fptr, fptrMember, hduNumMember, nGroupMembers
#</debug>

    def openAllMembers(self, fptr, filename, mode):
        """
        Opens all fits files that contain member HDUs.

        Files are opened readonly (mode=0) or readwrite (mode=1).

        Returns a list [(fptr1, hduNum1), (fptr2, hduNum2), ...] with the
        file pointers and the physical HDU numbers of the group members. 

        The routine operates on the current HDU.
        """
        
        status, nGroupMembers = \
                cfitsio.fits_get_num_members(fptr)
        if status:
            msg = "Error %d while opening group members in file %s" \
                  % (status, filename)
            myError = MBFITSError(msg)
            raise myError
        
        # Get the member's physical HDU numbers from the grouping table's
        # MEMBER_POSITION column:
        status, colnum = \
                cfitsio.fits_get_colnum(fptr, \
                                        1, \
                                        "MEMBER_POSITION")
        if status:
            msg = "Error %d while opening group members in file %s" \
                  % (status, filename)
            myError = MBFITSError(msg)
            raise myError

        status, pHduNumMembers = \
                cfitsio.fits_read_col_lng(fptr, \
                                          colnum, \
                                          1, \
                                          1, \
                                          nGroupMembers, \
                                          0)
        if status:
            msg = "Error %d while opening group members in file %s" \
                  % (status, filename)
            myError = MBFITSError(msg)
            raise myError
        
        # Open the members and get their file pointers at the same time:
        fptrMembers = []
        for iMember in xrange(1, nGroupMembers+1):
            status, fptrMember = \
                    cfitsio.fits_open_member(fptr, iMember)
            if status:
                msg = "Error %d while opening group member %d in file %s" \
                      % (status, iMember, filename)
                myError = MBFITSError(msg)
                raise myError

            fptrMembers.append(fptrMember)

        return zip(fptrMembers, pHduNumMembers)


#########################################################################
class Table:
#########################################################################
    
    def __init__(self,tableName='',Parent=None):

	""" Initialise a Table of type tableName.
	It is a child of 'Parent', which can be another Table or
	a FitsFile object. """

        # parent and child informations are written
        self.Parent=Parent 	# 'parent' table
        self.ChildTables=[] 	# list of tables related to this one
				#  - will be filled by creation of 'child'-tables

        if self.Parent != None:
            self.Parent.ChildTables.append(self)
	    self.HduMap = Parent.HduMap
            self.FitsFile = Parent.FitsFile
	else:
	    self.HduMap = None
            self.FitsFile = None

        self.Name=tableName

	# Create (initialise) one Header and one BinTable
        self.Header = Header(self)
        self.BinTable = BinTable(self)
        self.HduNum = -1 # This is the logic HduNum!



    def __str__(self):
	result = "Table name: "+self.Name
	result = result+"\n  ***  Header :\n"
	result = result+self.Header.__str__()
	result = result+"  ***  Bin Table :\n"
	result = result+self.BinTable.__str__()
	return result



    def _updateHduMaps(self, hduMap):

        """ Update HduMap in this table an all associated tables """
        
        self.HduMap = hduMap
        self.Header.HduMap = hduMap
        self.BinTable.HduMap = hduMap
        for Table in self.ChildTables:
            Table._updateHduMaps(self.HduMap)


    def getFilename(self):

        """
        Returns the name of the (physical) fits file in which the table
        resides.
        """

        filename = ""

        if self.HduNum>0 and self.HduMap:
            pFptr, pHdu = self.HduMap[self.HduNum]

            status, filename = cfitsio.fits_file_name(pFptr)
            if status:
                msg = 'Could not determine filename'
                myError = MBFITSError(msg,code=status)
                raise myError

        return filename
            
            

    def create(self, filename="", \
               SUBSNUM=None, FEBE=None, BASEBAND=None):
        """
        Create and write a new table in the file specified by filename.

        If filename is "", the new table is created in the file in which Parent
        resides.

        The other optional arguments are indexing information which is
        written into the FitsFile's grouping table.
        """

        if config.DEBUG:
            print 'creating table', self.Name
            
        if not self.HduMap:
            msg = 'No HduMap. File seems to be closed.'
            myError = MBFITSError(msg)
            raise myError

        if filename:
            # Try to open an existing file read-write:
            status, fptr = cfitsio.fits_open_file(filename, 1)
            if status:
                # Create it instead:
                status,fptr = cfitsio.fits_create_file(filename)
                if status:
                    msg = 'Could not create file', filename
                    myError = MBFITSError(msg,code=status)
                    raise myError
                status=cfitsio.fits_create_img(fptr,32,0,[])
                if status:
                    msg = 'Could not create primary image in file', filename
                    myError = MBFITSError(msg,code=status)
                    raise myError
                status, cpHduNum = cfitsio.fits_movabs_hdu(fptr, 1)
                if status:
                    msg = 'Could not move to primary image in file', filename
                    myError = MBFITSError(msg,code=status)
                    raise myError
                # Add MBFITSFILE keyword to primary header:
                if self.FitsFile:
                    key = "MBFITS"
                    val = self.FitsFile.currentFileName.split('/')[-1]
                    com = "Name of MBFitsFile"
                    status = cfitsio.fits_update_key_str(fptr, key, val, com)
                    if status:
                        msg = 'Could not update keyword %s in file %s' \
                              % (key, filename)
                        myError = MBFITSError(msg,code=status)
                        raise myError
        else:
            # Get the fptr from the FitsFile:
            fptr = self.FitsFile.fptr
        
        # get XML-Object corresponding to tableName
	tableName = self.Name
        table=XmlStruc.XmlScan.tables[tableName]
        # test if the correct 'parent' (ParentTable) is passed
        if self.Parent != None:
            if table.getAttribute('needsParentOfType')==self.Parent.Name:
                HasParent="true"
            else:
                msg = "For creation of table "+tableName
                msg = msg+" a parent of type "+table.getAttribute('needsParentOfType')
                msg = msg+" is needed (not "+self.Parent.Name+")."
                myError = MBFITSError(msg)
                raise myError
        else:
            HasParent="false"
            #if table.getAttribute('needsParentOfType')!='None':
                #raise ValueError, "For creation of table "+tableName+", a parent of \
		#	type "+table.getAttribute('needsParentOfType')+" is needed \
		#	(not "+ParentTable.name+")."

        # read table properties
        Type=int(table.getAttribute('type'))
        InitRows=int(table.getAttribute('initrows'))
        InitColumns = 0

        # create an empty FITS-table (will be filled after Keywords are filled)
        status = cfitsio.fits_create_tbl(fptr,Type,InitRows,InitColumns,[],[],[],tableName)

        # get physical HDU-Number
        status, pHduNum=cfitsio.fits_get_hdu_num(fptr)

        # add the table to the FitsFile:
        lHduNum = self.FitsFile.addTable(self, \
                                         fptr, \
                                         pHduNum, \
                                         EXTNAME=self.Name, \
                                         SUBSNUM=SUBSNUM, \
                                         FEBE=FEBE, \
                                         BASEBAND=BASEBAND)
#<test>
        self.HduNum = lHduNum
#</test>

        # get XML-Object corresponding to the header keywords of the table
        tblHeader=table.getElementsByTagName('HEADER')[0]

        # create Keywords from XML-Object
	self.Header.create(tblHeader.childNodes)

        # get XML-Object corresponding to the Columns of the table
        TblCols=table.getElementsByTagName('TABLE')[0]

        # get column descriptions from XML-Object
	self.BinTable.create(TblCols.childNodes)

        # create binary table columns, and update units and descriptions
        for i in xrange(len(self.BinTable.ttypes)):
            if self.BinTable.descriptions[i] !='':
                status = cfitsio.fits_insert_col(fptr,i+1,self.BinTable.ttypes[i],\
				self.BinTable.tforms[i])
                if self.BinTable.tdims[i] != '':
                    DimKey = 'TDim'+str(i+1)                        
                    status = cfitsio.fits_update_key(fptr,16,DimKey,self.BinTable.tdims[i],\
				'dimension of field')
                ttypeNr='TTYPE'+str(i+1)
                status = cfitsio.fits_update_key(fptr,16,ttypeNr,self.BinTable.ttypes[i],\
				self.BinTable.descriptions[i])
                status = cfitsio.fits_write_key_unit(fptr,ttypeNr,self.BinTable.units[i])
        if config.DEBUG:
            print '..... Table '+self.Name+' created.'



    def readHeader(self,HduNum):

        """ Read header of table with logical number HduNum """

        if not self.HduMap:
            msg = 'No HduMap. File seems to be closed.'
            myError = MBFITSError(msg)
            raise myError
        
	# Initialisations
        self.HduNum = HduNum

        # Physical file pointer and HduNum:
        fptr, pHduNum = self.HduMap[HduNum]
        
        status = cfitsio.fits_movabs_hdu(fptr,pHduNum)
        status, self.Name, comment = cfitsio.fits_read_keyword(fptr,'EXTNAME')
        
        # Remove the leading and trailing quotes
        self.Name = self.Name[1:-1]
	if config.DEBUG:
	    print "mbfits.Table.read: Now reading Table ",self.Name

        self.Header.read()


    def read(self,HduNum):

        """ Read table with logical number HduNum"""

        # First, read table header
        self.readHeader(HduNum)

        # Physical file pointer and HduNum:
        fptr, pHduNum = self.HduMap[HduNum]
        
	# Then, the column names, as well as the data types and array sizes of
	# the BinTable must be found in the relevant keywords of the header
	numFields = self.Header.KeywordDict['TFIELDS']
	numFields = int(numFields[0])
	if config.DEBUG:
	    print "Found TFIELDS = ",numFields
	for i in xrange(numFields):
	    keyType = 'TTYPE'+str(i+1)
	    keyForm = 'TFORM'+str(i+1)
	    keyDim = 'TDIM'+str(i+1)
	    tmp = self.Header.KeywordDict[keyType]
	    tt = tmp[0]
	    if tt[0] == "'":
	        tt = tt[1:-1]
	        tt.strip()
	    self.BinTable.ttypes.append(string.rstrip(tt))
	    tmp = self.Header.KeywordDict[keyForm]
	    tt = tmp[0]
	    if tt[0] == "'":
	        tt = tt[1:-1]
	        tt.strip()
	    self.BinTable.tforms.append(string.rstrip(tt))
	    if self.Header.KeywordDict.has_key(keyDim):
	        tmp = self.Header.KeywordDict[keyDim]
	        tt = tmp[0]
	        if tt[0] == "'":
	            tt = tt[1:-1]
	            tt.strip()
	        self.BinTable.tdims.append(string.rstrip(tt))
	    else:
	        self.BinTable.tdims.append('')

        # Read the Bin Table
        self.BinTable.read()



    def showObsmode(self):

        """ Prints the value of the keyword SCANTYPE if present """

        # FWY: this seems to be have changed into SCANTYPE
	if self.Header.KeywordDict.has_key('SCANTYPE'):
	    obsmode = self.Header.KeywordDict['SCANTYPE']
	    print "Found observing mode: ",repr(obsmode[0])



    def getObsmode(self):

        """ Returns the value of the keyword SCANTYPE if present. """
   
        try:
            result = self.Header.KeywordDict['SCANTYPE']
        except KeyError:
            result = None
        return result



    def getObstype(self):

        """ Returns the value of the keyword SCANTYPE if present.  """

        try:
	    result = self.Header.KeywordDict['SCANTYPE']
        except KeyError:
            result = None
        return result
        


    def getObsgeom(self):

        """ Returns the value of the keyword SCANGEOM if present. """

        try:
            result = self.Header.KeywordDict['SCANGEOM']
        except KeyError:
            result = None
        return result    



#########################################################################
class TblElement:
#########################################################################
    
    def __init__(self,XmlElement,parent):

        """ creates a MBFits header keyword or binary table column
	represented by a xmlElement """
        
        # read name, attributes and preset values for row content
        self.value      = str(XmlElement.getAttribute('preset'))
        self.type       = str(XmlElement.getAttribute('type'))
        self.descr      = str(XmlElement.getAttribute('description'))
        self.unit       = str(XmlElement.getAttribute('unit'))
        self.Parent     = parent
        self.XmlElement = XmlElement
        self.HduMap     = parent.HduMap
        self.FitsFile   = parent.FitsFile

        if XmlElement.hasChildNodes() == 1:            
            # search for the value of the tag (nodeType=3:Text) containing
	    # its keyword-name. Somewhat ugly, but I didn't find a method
	    # for direct access to the enclosed text
            for XmlNode in XmlElement.childNodes:
                if XmlNode.nodeType == 3:
                    self.name=str(XmlNode.data)
        else:
            msg = "Keyword not defined."
            myError = MBFITSError(msg)
            raise myError

        # <not needed>
        # get value via the 'connectorModule' module if a dataSource
	# is given by the XML file
        # ** we don't use dataSource in initialising the tables from the xml anymore
        # We use "preset='blah'" instead (see above).
        #if XmlElement.getAttribute('dataSource')!="":
        #    global connector
        #    ## ComClass = connector.ComClass()
	#    #print 'Designations:'
	#    #print ComClass.ScanObject.spatialSetup.pattern[0].designation
	#    #print ComClass.ScanObject.spatialSetup.pattern[1].designation
        #    DataSource = XmlElement.getAttribute('dataSource')
        #    Code = "TmpValue = "+DataSource
        #    try:
        #        exec Code
        #    except:
        #        print "WARNING: exec of code '%s' in mbfits.py raised an error. - Wasn't able to get the keyword's (%s) value."% (Code,self.name)
        #        TmpValue = ''
        #    self.value=TmpValue
        # </not needed>

        # convert datatypes
        if self.value:
            if string.find(self.type, 'A') >= 0:
                self.value=str(self.value)
            elif self.type in ['I','Idim','J','Jdim']:
                self.value=int(self.value)
            elif self.type in ['E','D','Edim','Ddim']:
                self.value=float(self.value)
        else:
            if self.type in ['I','Idim','J','Jdim','L','Ldim']:
                self.value=0
            elif self.type in ['E','D','Edim','Ddim']:                
                self.value=-0.
            # elif self.type == 'L':
            # self.value='F'

            

    def lookForChanges(self,value=None):
        
        """ Look if the keyword has changed in respect to it's
        'changeLookUp' attribute - returns 1 or 0 """

        if not self.HduMap:
            msg = 'No HduMap. File seems to be closed.'
            myError = MBFITSError(msg)
            raise myError
        
        if value:
            self.value = value
        lookUp = self.XmlElement.getAttribute('changeLookUp')
        if len(lookUp) >= 2:
               if lookUp[0]=="'" and lookUp[-1]=="'":
                   lookUp = string.strip(lookUp[1:-1])        
        dataSource = self.XmlElement.getAttribute('dataSource')
        if len(dataSource) >=2:
               if dataSource[0]=="'" and dataSource[-1]=="'":
                   dataSource = string.strip(dataSourceUp[1:-1])
            
        result = 1
        if lookUp != "":
            if (dataSource != "") or (dataSource == "" and value):
                lookUpTableName = lookUp.split('.')[0]
                lookUpKeyword = lookUp.split('.')[1]
                table=self.Parent.Parent
                lookUpTable = _searchTable(table,lookUpTableName)

                fptr, pHduNum = self.HduMap[lookUpTable.HduNum]
                status = cfitsio.fits_movabs_hdu(fptr, pHduNum)                
                lookUpValue = lookUpTable.Header.readKeyword(lookUpKeyword)
                lookUpValue = _setType(lookUpValue)

                fptr, pHduNum = self.HduMap[table.HduNum]
                status = cfitsio.fits_movabs_hdu(fptr,pHduNum)
                if type(lookUpValue) == type('string'):
                    if len(lookUpValue) >= 2:
                        if lookUpValue[0] == "'" and lookUpValue[-1] == "'":
                            lookUpValue = string.strip(lookUpValue[1:9])
                if lookUpValue == self.value:
                    result = 0
                else:
                    result = 1
            elif (dataSource == "") and not value:
                result = 0
            
        return result



#########################################################################
class Header:
#########################################################################
    
    def __init__(self,parent):
        self.Parent = parent
        self.Keywords = []
        self.KeywordDict = {}
        self.XMLHeaderKeywordList = None

	# copy the parent's HduMap as a new attribute
	self.HduMap = parent.HduMap
	self.FitsFile = parent.FitsFile
        


    def __str__(self):
	nb = len(self.KeywordDict)
	if nb == 0:
	    result = "No information found in Header\n"
	else:
	    keySort = self.KeywordDict.keys()
	    keySort.sort()
	    result = ""
	    for key in keySort:
		tmp = self.KeywordDict[key]
		result = result+"%8s%20s%52s\n" % (key,tmp[0],tmp[1])
	return result



    def create(self,XmlKeywordList):

        """ Create header keywords as defined in XML-Structure  """
        
        # Create Table Header Keywords
        # self.Parent = Parent	# already defined in the __init__ method

        if not self.HduMap:
            msg = 'No HduMap. File seems to be closed.'
            myError = MBFITSError(msg)
            raise myError
        
        self.XMLHeaderKeywordList = XmlKeywordList
        #for node in self.XMLHeaderKeywordList:
        #    if node.nodeType != 1:
        #        self.XMLHeaderKeywordList.remove(node)
                
#<timing test>
        t0 = time.time()
        global dif2, dif3, dif5
        dif2 = 0
        dif3 = 0
        dif5 = 0
        sum01 = 0.
        sum12 = 0.
        sum3 = 0.
#</timing test>

        for XmlKeyword in self.XMLHeaderKeywordList:
            # search for elements (nodeType=1:Element) and create
	    # header keyword entrys in the fits-file
            if XmlKeyword.nodeType == 1:

#<timing test>
                t10 = time.time()
#</timing test>
                
                Element = TblElement(XmlKeyword,self)

#<timing test>
                t11 = time.time()
#</timing test>

                if Element.lookForChanges():
                    # For not writing all the status messages from updateKeyword,
                    # the debug mode is set to 0
                    debugValue = config.DEBUG
                    config.DEBUG = 0
#<timing test>                    
                    t31 = time.time()
#</timing test>
                    self.updateKeyword(Element.name,Element.unit,Element.value,descr=Element.descr)
#<timing test>
                    t32 = time.time()
                    sum3 += t32-t31
#</timing test>
                    # and debug mode being set back to it's original value
                    config.DEBUG = debugValue

#<timing test>
                t12 = time.time()

                sum01 += t11-t10
                sum12 += t12-t11
#</timing test>

#<timing test>
        t1 = time.time()

        print "RS_TIMING Header.create #keywords:", len(XmlKeywordList)
        print ""
        print "RS_TIMING t11-t10: ", sum01, "\tTblElement"
        print "RS_TIMING t12-t11: ", sum12, "\tupdateKeyword"
        print "CK_TIMING t32-t31: ", sum3 , "\tdavon updateKeyword Methode"
        print "" 
        print "RS_TIMING t1-t0: ", t1-t0
        print ""
        print "updateKeyword RESULTS"
        print
        print "dif2:",dif2,"\tfirst XML Loop"
        print "dif3:",dif3,"\tsecond XML Loop"
        print "dif5:",dif5,"\tdata writing"
        print
#</timing test>



    def addKeyword(self,key,unit,value,comment):

        """ Move to table's HDU number and write a new keyword entry.
        This does not require that the keyword was in the XML!
        The XML is checked for matches - and if found it calls updateKeyword()
        Otherwise it determines the data type from the value."""


        if not self.HduMap:
            msg = 'No HduMap. File seems to be closed.'
            myError = MBFITSError(msg)
            raise myError
        
        # check the xml
        if not self.XMLHeaderKeywordList:
            # get the XML description for this type of table header
            table     = XmlStruc.XmlScan.tables[self.Parent.Name]
            tblHeader = table.getElementsByTagName('HEADER')[0]
            self.XMLHeaderKeywordList = tblHeader.childNodes

        # get the XML element corresponding to the keyword
        found = None
        for XmlKeyword in self.XMLHeaderKeywordList:
            if XmlKeyword.nodeType == 1:
                for XmlNode in XmlKeyword.childNodes:
                    if XmlNode.nodeType == 3:
                        name=str(XmlNode.data)
                        if name == key:
                            Element = TblElement(XmlKeyword,self)
                            found = 1
                            break

        # If the keyword is in the xml then use updateKeyword
        if found and Element.lookForChanges(value):
            self.updateKeyword(key,unit,value,comment)
            return

        # Otherwise, add it anyway...
        
        # Determines the datatype; typeCode is needed by fits_write_key
        if (type(value) == str):
            typeCode = 16
            if ((value == 'F') or (value == 'T')):
                typeCode = 14
        elif (type(value) == float):
            typeCode = 82
        elif (type(value) == int):
            typeCode = 21
            
        # Physical file pointer and HduNum:
        fptr, pHduNum = self.HduMap[self.Parent.HduNum]
        
        status = cfitsio.fits_movabs_hdu(fptr,pHduNum)
        status = cfitsio.fits_write_key(fptr,typeCode,key,value,comment)
        status = cfitsio.fits_write_key_unit(fptr,key,unit)
        if config.DEBUG:
            print '.......... Keyword '+key+' added to table '+self.Parent.Name+'. value = ',value



    def addComment(self, comment):

        """ Add  a comment to the table"""

        if not self.HduMap:
            msg = 'No HduMap. File seems to be closed.'
            myError = MBFITSError(msg)
            raise myError
        
        # Physical file pointer and HduNum:
        fptr, pHduNum = self.HduMap[self.Parent.HduNum]
        
        status = cfitsio.fits_write_comment(fptr,str(comment))



    def updateKeyword(self,Key,Unit,Value,descr=None):

        """ Move to table's HDU number and update a keyword entry. """

        if not self.HduMap:
            msg = 'No HduMap. File seems to be closed.'
            myError = MBFITSError(msg)
            raise myError

        if not self.XMLHeaderKeywordList:
            # get the XML description for this type of table header
            table=XmlStruc.XmlScan.tables[self.Parent.Name]
            tblHeader=table.getElementsByTagName('HEADER')[0]
            self.XMLHeaderKeywordList = tblHeader.childNodes
#<timing test>
        global dif2, dif3, dif5
        t01 = time.time()
#</timing test>
        # get the XML element corresponding to the keyword
        found = None
        xmlKeyType = None
        for XmlKeyword in self.XMLHeaderKeywordList:
            if XmlKeyword.nodeType == 1:
                for XmlNode in XmlKeyword.childNodes:
                    try:
                        name=str(XmlNode.data)
                        if name == Key:
                            Element = TblElement(XmlKeyword,self)
                            xmlKeyType = str(XmlKeyword.getAttribute('type'))
                            found = 1
                            break
                    except:
                        pass
                if found:
                    break
#<timing test>
        t02 = time.time()
        dif2 += t02-t01       
#</timing test>
        if found and Element.lookForChanges(Value):
            if not xmlKeyType:
                OrigValue = ''
                OrigComment = ''
                #xmlKeyType = None   # useful to test if the required Key exists
                                # in the XML description of this table

                # Get the table's XML description
                XmlTable = XmlStruc.XmlScan.tables[self.Parent.Name]
                # Get the header's XML description
                XmlHeader = XmlTable.getElementsByTagName('HEADER')[0]
                # Loop over all keywords
                for XmlElement in XmlHeader.childNodes:
                    for XmlNode in XmlElement.childNodes:
                        if XmlNode.nodeType == 3:
                            if Key == str(XmlNode.data):
                                xmlKeyType = str(XmlElement.getAttribute('type'))
                                break
                    if xmlKeyType:
                        break
#<timing test>
            t03 = time.time()
            dif3 += t03-t02
#</timing test>

            if xmlKeyType:
                if ('A' in xmlKeyType):
                    keyType = str
                    typeCode = 16
                elif ('E' in xmlKeyType):
                    keyType = float
                    typeCode = 42
                elif ('D' in xmlKeyType):
                    keyType = float
                    typeCode = 82
                elif ('I' in xmlKeyType):
                    keyType = int
                    typeCode = 21
                elif ('J' in xmlKeyType):
                    keyType = int
                    typeCode = 41            
                elif ('L' in xmlKeyType):
                    keyType = int
                    typeCode = 14

                if (type(Value) == keyType) or ((type(Value)==long) and (keyType ==int)):
                    # Physical file pointer and HduNum:
#<debug>
##                     print "RS_DEBUG Header.updateKeyword self.Parent.HduNum:", self.Parent.HduNum
#</debug>
                    fptr, pHduNum = self.HduMap[self.Parent.HduNum]
                    status = cfitsio.fits_movabs_hdu(fptr, pHduNum)
                    status, OrigValue, OrigComment = cfitsio.fits_read_keyword(fptr,Key)
                    if descr:
                        OrigComment = str(descr)
#<timing test>
                    t04 = time.time()
#</timing test>

                    # select the right fits_update_key_TYPE function
                    if ('L' in xmlKeyType):
                        status = cfitsio.fits_update_key_log(fptr,Key,Value,OrigComment)
                    elif ('A' in xmlKeyType):
#<test>
                        # Truncate Value if necessaary:
                        pos = xmlKeyType.find('A')
                        if pos:
                            if xmlKeyType[:pos].isdigit():
                                maxlength = int(xmlKeyType[:pos])
                                if len(Value) > maxlength:
                                    print "RS_DEBUG: Keyword %s Value %s truncated" \
                                          % (Key, Value)
                                Value = Value[:maxlength]
#</test>
                        status = cfitsio.fits_update_key_str(fptr,Key,Value,OrigComment)
                    elif ('E' in xmlKeyType):
                        # Set number of decimals to -7 as defined in the generic function
                        # for updating float keys in cfitsio
                        decimals = -7
                        status = cfitsio.fits_update_key_flt(fptr,Key,Value,decimals,OrigComment)
                    elif ('D' in xmlKeyType):
                        # Set number of decimals to -15 as defined in the generic function
                        # for updating double keys in cfitsio
                        decimals = -15
                        status = cfitsio.fits_update_key_dbl(fptr,Key,Value,decimals,OrigComment)
                    elif ('I' in xmlKeyType):
                        status = cfitsio.fits_update_key_lng(fptr,Key,Value,OrigComment)
                    elif ('J' in xmlKeyType):
                        status = cfitsio.fits_update_key_lng(fptr,Key,Value,OrigComment)
#<timing test>
                    t05 = time.time()
                    dif5 += t05-t04
#</timing test>

                    status = cfitsio.fits_write_key_unit(fptr,Key,Unit)
                    if config.DEBUG:
                        print '.......... Keyword '+Key+' updated in table ',
                        print self.Parent.Name+'. value = ',Value

                else:
                    msg = 'Key '+Key
                    msg = msg+' can not be written to Table '
                    msg = msg+self.Parent.Name+', because of wrong type '
                    msg = msg+'(needs to be '+str(keyType)+', not '+str(type(Value))+'.'
                    myError = MBFITSError(msg)
                    raise myError
            else:
                if config.DEBUG:
                    msg = 'Key '+Key+' not found in the XML description of Table '
                    msg = msg+self.Parent.Name+"\n... just ignoring it!..."
                    print msg



#    def updateKeyword_lng(self,Key,Unit,Value):
#
#        """ Move to table's HDU number and update a keyword
#	(type=long) entry. """
#        
#	fptr = self.fptr
#        OrigValue = ''
#        OrigComment = ''
#        
#        status, OrigValue, OrigComment = cfitsio.fits_read_keyword(fptr,Key)
#        status = cfitsio.fits_movabs_hdu(fptr,self.Parent.HduNum)
#        pcfitsio.fits_modify_key_lng(fptr,Key,Value,OrigComment)
#        pcfitsio.fits_write_key_unit(fptr,Key,Unit)
#        print '.......... Keyword '+Key+' updated in table ',
#	print self.Parent.Name+'. value = ',Value


#    def updateKeyword_log(self,Key,Unit,Value):
#
#        """ Move to table's HDU number and update a keyword 
#	(type=logical) entry. """
#
#	fptr = self.fptr
#        OrigValue = ''
#        OrigComment = ''
#        
#        OrigValue, OrigComment = pcfitsio.fits_read_keyword(fptr,Key)
#        pcfitsio.fits_movabs_hdu(fptr,self.Parent.HduNum)
#        pcfitsio.fits_modify_key_log(fptr,Key,Value,OrigComment)
#        pcfitsio.fits_write_key_unit(fptr,Key,Unit)
#        print '.......... Keyword '+Key+' updated in table ',
#	print self.Parent.Name+'. value = ',Value



    def read(self):

        """ Read header keywords from MBFits-File  """
        
	# Use commands from the fitsio.read_keys method to read the header
	# Avoid using the Xml description, read all keywords

        if not self.HduMap:
            msg = 'No HduMap. File seems to be closed.'
            myError = MBFITSError(msg)
            raise myError

        #print "RS_DEBUG Header.read self.Parent.HduNum:", self.Parent.HduNum
        
        # Physical file pointer and HduNum:
        fptr, pHduNum = self.HduMap[self.Parent.HduNum]

        print "RS_DEBUG Header.read fptr, pHduNum:", fptr, pHduNum 


	result = {}
	status, num, morekeys = cfitsio.fits_get_hdrspace(fptr)
        for i in xrange(1,num+1):
            status, record = cfitsio.fits_read_record(fptr,i)
            if type(record) == type('string'):
                thiskey = string.strip(record[0:8])
            if (thiskey != 'COMMENT'):
                status, value, comment = cfitsio.fits_read_keyword(fptr,thiskey)
                result[thiskey] = [value,comment]
                self.Keywords.append(thiskey)
            
	for key in result.keys():
            result[key][0] = _setType(result[key][0])

        self.KeywordDict = result



    def readKeyword(self,key):

        """ Reads a keyword entry of the current HDU. """

        if not self.HduMap:
            msg = 'No HduMap. File seems to be closed.'
            myError = MBFITSError(msg)
            raise myError
        
        # Physical file pointer and HduNum:
        fptr, pHduNum = self.HduMap[self.Parent.HduNum]

        status = cfitsio.fits_movabs_hdu(fptr,pHduNum)
        
        key = str(key)
        status, value, comment = cfitsio.fits_read_keyword(fptr,key)
        return value



#########################################################################
class BinTable:
#########################################################################

    def __init__(self,parent):
        self.Parent=parent
	self.HduMap = parent.HduMap
	self.FitsFile = parent.FitsFile

	# Initialise attributes: description of table columns

        # Fixme: The following lists are not treated properly
        #        See methods _getTTYPE, _getTFORM, _getTDIM,
        #        _getDataDimension, and updateDimension
        self.ttypes=[]
        self.tforms=[]
        self.descriptions=[]
        self.units=[]
        self.tdims=[]
        # /Fixme

	# And a "Data" attribute, where the data can be stored
        self.Data = {}
        self._writeColFunctions = []
        self._readColFunctions = []

        # Mapping of FITSIO Datatypes to Numeric Typecodes for each column:
        self._colNumericTypecodes = []

    def __str__(self):
	if len(self.ttypes) == 0:
	    result = 'No table column found\n'
	else:
	    result = str(len(self.ttypes)) + ' columns found:\n'
	    for name in self.ttypes:
		result = result+'  .. '+name+'\n'
	if len(self.Data) == 0:
	    result = result+'No data found\n'
	else:
	    result = result+'Size of data: \n'
            for key in self.Data.keys():
                result = result+key+' -> '+str(len(self.Data[key]))+'\n'
                result = result+str(type(self.Data[key]))+'\n'
                if (type(self.Data[key]) == list):
                    result = result+str(type(self.Data[key][0]))+'\n'
                else:
                    result = result+str(self.Data[key].typecode())+'\n'

	return result



    def _createWriteColFunctions(self):

        """ Adds the appropriate cfitsio fits_write_col_TYPE function to the
        _writeColFunctions list, so it isn't needed to check for the datatype
        when adding a column entry"""
        
        # Physical file pointer and HduNum:
        fptr, pHduNum = self.HduMap[self.Parent.HduNum]

        status, numCols = cfitsio.fits_get_num_cols(fptr)
        for colNum in xrange(1,numCols+1):
            colType = self.Parent.Header.readKeyword('TFORM'+str(colNum))
            if 'L' in colType:
                self._writeColFunctions.append(cfitsio.fits_write_col_log)
            elif 'A' in colType:
                self._writeColFunctions.append(cfitsio.fits_write_col_str)
            elif ('E' in colType):
                self._writeColFunctions.append(cfitsio.fits_write_col_flt)
            elif ('D' in colType):
                self._writeColFunctions.append(cfitsio.fits_write_col_dbl)
            elif ('I' in colType):
                self._writeColFunctions.append(cfitsio.fits_write_col_int)
            elif ('J' in colType):
                self._writeColFunctions.append(cfitsio.fits_write_col_lng)
            else:
                msg = 'Type'+colType
                msg = msg+'can not be written to Table'
                msg = msg+self.Parent.Name+'.'
                myError = MBFITSError(msg)
                raise myError



    def _createReadColFunctions(self):
        
        """ Adds the appropriate cfitsio fits_read_col_TYPE function to the
        _readColFunctions list, so it isn't needed to check for the datatype
        when reading a column entry"""
        
        # Physical file pointer and HduNum:
        fptr, pHduNum = self.HduMap[self.Parent.HduNum]

        status, numCols = cfitsio.fits_get_num_cols(fptr)
        for colNum in xrange(1,numCols+1):
            colType = self.Parent.Header.readKeyword('TFORM'+str(colNum))
            if 'L' in colType:
                self._readColFunctions.append(cfitsio.fits_read_col_log)
            elif 'A' in colType:
                self._readColFunctions.append(cfitsio.fits_read_col_str)
            elif ('E' in colType):
                self._readColFunctions.append(cfitsio.fits_read_col_flt)
            elif ('D' in colType):
                self._readColFunctions.append(cfitsio.fits_read_col_dbl)
            elif ('I' in colType):
                self._readColFunctions.append(cfitsio.fits_read_col_int)
            elif ('J' in colType):
                self._readColFunctions.append(cfitsio.fits_read_col_lng)
            else:
                msg = 'Type'+colType
                msg = msg+'can not be read from Table'
                msg = msg+self.Parent.Name+'.'
                myError = MBFITSError(msg)
                raise myError



    def _getColDimensions(self,Element):
        
            # correct type has to be created from XML informations
            # get datatype (first character (0))
            ArrayType=Element.type[0]
            # create list from element.type definition
            TypeList=string.split(Element.type[string.find(Element.type,\
                        "(")+1:string.rfind(Element.type,")")],",")
 
            if ArrayType == 'A':
                strLength = int(TypeList[-1])
                TypeList = TypeList[:-1]

            # create string containing the dimension information for
            #  self.tdims[] and fill with values from (other) table headers
            ArrayDim='('

            # for every header keyword dependency get the the value
            # from the corresponding table
            for Type in TypeList:
                if (string.find(Type,'(')>0):

                    # extract TableName from XML type description
                    TableName = Type[0:string.find(Type,"(")]
                    # extract header keyword from XML type description
                    TableHeaderKeyword = Type[string.find(Type,"(")+1:string.rfind(Type,")")]
                    # get table referred to by TableName
                    Table = _searchTable(self.Parent,TableName)

                    # set HDU to referred table, get value for the header keyword set HDU back
                    fptr, pHduNum = self.HduMap[Table.HduNum]
                    status = cfitsio.fits_movabs_hdu(fptr,pHduNum)
                    status, value, comment = cfitsio.fits_read_keyword(fptr,TableHeaderKeyword)
                    TypeVal = str(value)
                    # cut everything right from a decimal point (only INT is allowed)
                    if string.rfind(TypeVal,'.') >0:
                        TypeVal = str(TypeVal[:string.rfind(TypeVal,'.')])
                    ArrayDim = ArrayDim+TypeVal+","

                    #Array = Array+str(2)+","
                    fptr, pHduNum = self.HduMap[self.Parent.HduNum]
                    status = cfitsio.fits_movabs_hdu(fptr,pHduNum)
                else:
                    ArrayDim = ArrayDim+str(Type)+','

            ArrayDim=ArrayDim[0:len(ArrayDim)-1]+')'
            ArrayList=string.split(ArrayDim[string.find(ArrayDim,"(")+1:string.rfind(ArrayDim,")")],",")
            # calculate length of datafield
            TypeLength = 1
            for Dimension in ArrayList:
                if Dimension.isdigit():
                    TypeLength = TypeLength * int(Dimension)
            if ArrayType == 'A':
                TypeLength *= strLength
                # if the string is multidimensional then we need to use the format e.g. '10A2'
                # so add the individual string lengths to the end
                if TypeLength != strLength:
                    ArrayType += str(strLength)

            strArrayType = str(TypeLength) + str(ArrayType)
            
            return ArrayDim,ArrayList,strArrayType



    def create(self,XmlColumnList):

        """ Create binary table columns as defined in XML-Structure  """

        if not self.HduMap:
            msg = 'No HduMap. File seems to be closed.'
            myError = MBFITSError(msg)
            raise myError
        
        # Create List of Table Columns from the Xml description
        for XmlColumn in XmlColumnList:
            if XmlColumn.nodeType==1:
                Element=TblElement(XmlColumn,self)
                # testing on dependencies on keywords (if type is 'Xdim')
                if string.find(Element.type,"(")>=0 and string.find(Element.type,"1P")<0:
                    # get the description for the 'TDIMx' keyword
                    ArrayDim,ArrayList,strArrayType = self._getColDimensions(Element)

                    # TTYPEn keywords give the names of the fields, and TFORMn give the types of the data
                    self.tforms.append(string.rstrip(strArrayType))

                    # check if array is multidimensional or not                                
                    if len(ArrayList) > 1:
                        self.tdims.append(string.rstrip(str(ArrayDim)))
                    else:
                        #                        ArrayDim = '(1,'+str(ArrayList[0])+')'
                        #                        self.tdims.append(ArrayDim)
                        self.tdims.append('')
                        
                else:
                    self.tforms.append(string.rstrip(str(Element.type)))
                    self.tdims.append('')
                self.ttypes.append(string.rstrip(str(Element.name)))
                self.units.append(str(Element.unit))
                self.descriptions.append(str(Element.descr))



    def _getTFORM(self,colNum):

        """ Get the dimension and datatype description for this column (colNum).
        Returns the TFORM# keyword string."""
        
        # Fixme: This function is broken! See updateDimension


        # Handle non-scalar data: read the TFORMi keyword
        try:
            tform = self.tforms[colNum-1]
        # if this fails, get it from the file
        except IndexError:                
            # Physical file pointer and HduNum:
            fptr, pHduNum = self.HduMap[self.Parent.HduNum]

            status, tform, comment = cfitsio.fits_read_keyword(fptr,'TFORM'+str(colNum))
            if not status:
                # Remove leading and trailing quote in the tform string
                if len(tform) > 0:
                    if tform[0] == "'":
                        tform = tform[1:-1]
                    tform = string.rstrip(tform)
                self.tforms.append(tform)
            else:
                tform = ''
        return tform



    def _getTTYPE(self,colNum):

        """ Get name of this column (colNum).
        Returns the TTYPE# keyword string."""
        
        # Fixme: This function is broken! See updateDimension

        try:
            ttype = self.ttypes[colNum-1]
        # if this fails, get it from the file
        except IndexError:                
            # Physical file pointer and HduNum:
            fptr, pHduNum = self.HduMap[self.Parent.HduNum]

            status, ttype, comment = cfitsio.fits_read_keyword(fptr,'TTYPE'+str(colNum))
            if not status:
                # Remove leading and trailing quote in the tform string
                if len(ttype) > 0:
                    if ttype[0] == "'":
                        ttype = ttype[1:-1]
                    ttype = string.rstrip(ttype)
                self.ttypes.append(ttype)
            else:
                ttype = ''
        return ttype



    def _getTDIM(self,colNum):

        """ Get dimension information if multidimensional for ccolumn (colNum).
        Returns the TDIM# keyword string."""
        
        # Fixme: This function is broken! See updateDimension

        try:
            tdim = self.tdims[colNum-1]
        # if this fails, get it from the file
        except IndexError:                
            # Physical file pointer and HduNum:
            fptr, pHduNum = self.HduMap[self.Parent.HduNum]

            status, tdim, comment = cfitsio.fits_read_keyword(fptr,'TDIM'+str(colNum))
            if not status:
                # Remove leading and trailing quote in the tform string
                if len(tdim) > 0:
                    if tdim[0] == "'":
                        tdim = tdim[1:-1]
                    tdim = string.rstrip(tdim)
                self.tdims.append(tdim)
            else:
                tdim = ''
        return tdim



    def _getDataDimension(self,colType):

        """ Extract the dimension from the TFORM# keyword string (colType description)."""
        
        # Fixme: This function is broken! See updateDimension

        # Then check if a number (means multidimensional if >1) is present in TFORMi, except for strings
        if string.find(colType,'A') < 0:
            if colType[0] in ['1','2','3','4','5','6','7','8','9']:
                if colType[1] != 'P':
                    dataDim = int(colType[0:-1])
                # Special handling for varaiable length arrays
                else:
                    dataDim = int(colType[string.find(colType,'(')+1:-1])
            else:
                dataDim = 1
        else:
            dataDim = 1
            
        return dataDim



#    def addTableEntry(self,ColName,DataList):
#
#        """ Move to table's HDU number and write a new table entry. """
#
#	fptr = self.fptr
#        status = cfitsio.fits_movabs_hdu(fptr,self.Parent.HduNum)
#        status, ColNum = cfitsio.fits_get_colnum(fptr,0,ColName)[1]
#        status, numRows = cfitsio.fits_get_num_rows(fptr)
#        status = cfitsio.fits_write_col(fptr,ColNum,numRows+1,1,DataList)
#        if config.DEBUG:
#	    print '.......... Entry added to table ',
#	    print self.Parent.Name+', column '+ColName+'.'



    def updateDimension(self,colName):

        # Fixme: This function is broken!
        #        It must update the lists
        #          self.ttypes
        #          self.tforms
        #          self.tdims
        #        but it doesn't. Therefore, the methods _getTTYPE,
        #        _getTFORM, _getTDIM, _getDataDimension, and updateDimension
        #        are broken as well!

        if not self.HduMap:
            msg = 'No HduMap. File seems to be closed.'
            myError = MBFITSError(msg)
            raise myError
        
        # Physical file pointer and HduNum:
        fptr, pHduNum = self.HduMap[self.Parent.HduNum]

        status = cfitsio.fits_movabs_hdu(fptr,pHduNum)
        
        # get the XML description for this table
        # and get the dimensions for this column
        XmlTable=XmlStruc.XmlScan.tables[self.Parent.Name]
        # get XML-Object corresponding to the Columns of the table
        XmlTblCols=XmlTable.getElementsByTagName('TABLE')[0]
        i = 0
        for XmlColumn in XmlTblCols.childNodes:
            if XmlColumn.nodeType==1:
                i += 1
                for XmlNode in XmlColumn.childNodes:
                    if XmlNode.nodeType == 3:
                        XmlColName = str(XmlNode.data)
                Element=TblElement(XmlColumn,self)
                if XmlColName == colName or colName == 1:
                    ArrayDim,ArrayList,strArrayType = self._getColDimensions(Element)

                    status = cfitsio.fits_movabs_hdu(fptr,pHduNum)
                    status, OrigValue, OrigComment = cfitsio.fits_read_keyword(fptr,'TDIM'+str(i))
                    # for TDIM only update columns that already have a TDIM keyword
                    if not status:
                        status = cfitsio.fits_update_key_str(fptr,'TDIM'+str(i),ArrayDim,OrigComment)
                    # but update all vector lengths..
                    status, colNum = cfitsio.fits_get_colnum(fptr,1,XmlColName)
                    if 'A' in strArrayType:
                        vectorLength = int(strArrayType.split('A')[0])
                    else:
                        vectorLength = int(strArrayType[:-1])
                    status = cfitsio.fits_modify_vector_len(fptr,colNum,vectorLength)



    def setVariableLengthDimension(self,colName,length,length2=0):

        """ Method to manually set the dimension of a variable-length array column.
            If length2 is specified then it will get a TDIM (except for strings)
        """
        
        if not self.HduMap:
            msg = 'No HduMap. File seems to be closed.'
            myError = MBFITSError(msg)
            raise myError

        # Physical file pointer and HduNum:
        fptr, pHduNum = self.HduMap[self.Parent.HduNum]

        status = cfitsio.fits_movabs_hdu(fptr,pHduNum)
        status, colNum = cfitsio.fits_get_colnum(fptr,1,colName)
        key = 'TFORM'+str(colNum)
        status, OrigValue, OrigComment = cfitsio.fits_read_keyword(fptr,key)
        ArrayDim = 'none'
        if 'P' in OrigValue:
            # get the actual string e.g. '1PE'
	    # The string contains the "'" characters, thus [1:-1]
            value = OrigValue[1:-1].split('(')[0]
            value = value.rstrip()
            # If multi-dimensional then fill TDIM (except for strings):
            if length2:
                if 'A' in OrigValue:
                    # string cannot be multidimensional => must join all together
                    value = value + '(' + str(length*length2) + ')'
                else:
                    value = value + '(' + str(length*length2) + ')'
                    ArrayDim = '('+str(length)+","+str(length2)+")"
                    status = cfitsio.fits_update_key_str(fptr,'TDIM'+str(colNum),ArrayDim,'dimension of field')
            # otherwise update the '1Px' with length in brackets e.g. '1PE(20)'
            else:
                value = value + '(' + str(length) + ')'
            # update the TFORM:
            status = cfitsio.fits_update_key_str(fptr,key,value,OrigComment)
	    print 'DEBUG: setVari status=', status
            
            if config.DEBUG:
                print '.......... Variable-length array dimension for column %s of table %s set to %s with TDIM %s'%(colName,self.Parent.Name,value,ArrayDim)
        else:
            print '%s not a variable-length column - ignoring command to set variable-length dimension'%(colName)



    def addTableRow(self,DataList):

        """ Move to table's HDU number and write a new table row. """

        if not self.HduMap:
            msg = 'No HduMap. File seems to be closed.'
            myError = MBFITSError(msg)
            raise myError
        
        # Physical file pointer and HduNum:
        fptr, pHduNum = self.HduMap[self.Parent.HduNum]

        status = cfitsio.fits_movabs_hdu(fptr,pHduNum)

        # create the set of functions needed to write one column
        # add the function needed for the type of the actual column

        if len(self._writeColFunctions) == 0:
            self._createWriteColFunctions()
            
        # for every dataList item, one column after the other is filled
        status, numCols = cfitsio.fits_get_num_cols(fptr)
        if numCols == len(DataList):
            
            colNum=1
            status, numRows = cfitsio.fits_get_num_rows(fptr)
            # add one new (empty) row
            status = cfitsio.fits_insert_rows(fptr,numRows,1)
            for Item in DataList:
                if type(Item)!=list and \
                   type(Item)!=type(Numeric.zeros(1)):
                    Item = [Item]
                status = self._writeColFunctions[colNum-1](fptr,colNum,numRows+1,1,len(Item),Item)
                colNum = colNum+1
            if config.DEBUG:
	        print '.......... Row added to table '+self.Parent.Name+'.'
        else:
            msg = "Number of values ("+str(len(DataList))
            msg = msg+") not corresponding to number of columns ("
            msg = msg+str(numCols)+")."
            myError = MBFITSError(msg)
            raise myError



    def addTableRows(self,DataList):

        """ Move to table's HDU number and write a set of table rows. """

        
	print 'DEBUG: ADDTABLEROWS NAME:', self.Parent.Name, 'HDUNUM:', self.Parent.HduNum

        if not self.HduMap:
            msg = 'No HduMap. File seems to be closed.'
            myError = MBFITSError(msg)
            raise myError

        # Physical file pointer and HduNum:
        fptr, pHduNum = self.HduMap[self.Parent.HduNum]

        status = cfitsio.fits_movabs_hdu(fptr,pHduNum)

        # create the set of functions needed to write one column
        # add the function needed for the type of the actual column

        if len(self._writeColFunctions) == 0:
            self._createWriteColFunctions()
                    
        # for every dataList item, one column after the other is filled
        status, numCols = cfitsio.fits_get_num_cols(fptr)
        if numCols == len(DataList):
            colNum=1
            status, numRows = cfitsio.fits_get_num_rows(fptr)

            calls = 0
            for Item in DataList:
                if type(Item) != list and \
                   type(Item)!=type(Numeric.zeros(1)):
                    msg = 'To add multiple rows, the DataList needs to contain Lists or Numeric.arrays'
                    myError = MBFITSError(msg)
                    raise myError                

                rowNum = numRows+1
                if type(Item[0])!=list and \
                   type(Item[0])!=type(Numeric.zeros(1)):
                    lenItem = len(Item)                        
                    status = self._writeColFunctions[colNum-1](fptr,colNum,rowNum,1,lenItem,Item)
                    calls = calls+1
                else:
                    for rowData in Item:
                        if type(rowData)==list or \
                           type(rowData)==type(Numeric.zeros(1)): 
                            lenRowData = len(rowData)
                        else:
                            lenRowData = 1
                        #print "colNum = ",colNum," rowNum = ",rowNum
                        status = self._writeColFunctions[colNum-1](fptr,\
                                                         colNum,rowNum,1,lenRowData,rowData)
                        calls=calls+1
                        rowNum = rowNum+1
                rowsAdded = len(Item)
                    
                colNum = colNum+1
            if config.DEBUG:
	        print '.......... '+str(rowsAdded)+' rows added to table '+self.Parent.Name+'. '
        else:
            msg = "Number of values ("+str(len(DataList))
            msg = msg+") not corresponding to number of columns ("
            msg = msg+str(numCols)+")."
            myError = MBFITSError(msg)
            raise myError


    def getOptNumRowsToAdd(self):
        """
        Returns the optimal number of rows to read or write at one time for
        maximum efficiency.

        See cfitsio manual for details of fits_get_rowsize
        """

        if not self.HduMap:
            msg = 'No HduMap. File seems to be closed.'
            myError = MBFITSError(msg)
            raise myError

        # Physical file pointer and HduNum:
        fptr, pHduNum = self.HduMap[self.Parent.HduNum]

        status = cfitsio.fits_movabs_hdu(fptr,pHduNum)

        status, nrows = cfitsio.fits_get_rowsize(fptr)
        return nrows

    def _createTypecodes(self):

        """
        Adds an entry with the appropraite Numeric typecode
        to self._colNumericTypecodes for each column.

        According to the cfitsio documentation, the mapping of
        FITSIO datatypes to Numeric Typecodes might be machine dependent!
        """

        def _cannotHandleType(colType):
            msg = 'Type'+colType
            msg = msg+'can not be handled in Table'
            msg = msg+self.Parent.Name+'.'
            myError = MBFITSError(msg)
        
        # Physical file pointer and HduNum:
        fptr, pHduNum = self.HduMap[self.Parent.HduNum]

        status, numCols = cfitsio.fits_get_num_cols(fptr)
        for colNum in xrange(1,numCols+1):
            colType = self.Parent.Header.readKeyword('TFORM'+str(colNum))
            if 'L' in colType:
                self._colNumericTypecodes.append(Numeric.Int0)
            elif 'B' in colType:
                # 'B', 'SB'
                self._colNumericTypecodes.append(Numeric.Int0)
            elif ('I' in colType):
                # 'I', 'UI'
                self._colNumericTypecodes.append(Numeric.Int16)
            elif ('K' in colType):
                # 'K', 'UK'
                self._colNumericTypecodes.append(Numeric.Int32)
            elif ('J' in colType):
                # 'J', 'JJ', 'UJ'
                self._colNumericTypecodes.append(Numeric.Int32)
            elif ('E' in colType):
                self._colNumericTypecodes.append(Numeric.Float32)
            elif ('F' in colType):
                if ('FC' in colType):
                    _cannotHandleType(colType)
                elif ('FM' in colType):
                    _cannotHandleType(colType)
                else:
                    self._colNumericTypecodes.append(Numeric.Float32)
            elif ('D' in colType):
                self._colNumericTypecodes.append(Numeric.Float64)
            elif ('G' in colType):
                self._colNumericTypecodes.append(Numeric.Float64)
            else:
                _cannotHandleType(colType)

    def addTableRowsDatapar(self, naData):

        """
        Specialized method to fill the DATAPAR table.

        naData is a Numeric.array of shape (numRowsData, numColsData).

        The slices of naData for individual columns will be casted to the
        datatype as defined in the XML file.
        """

        if not self.HduMap:
            msg = 'No HduMap. File seems to be closed.'
            myError = MBFITSError(msg)
            raise myError
        
## 	print 'DEBUG: ADDTABLEROWSDATAPAR NAME:', self.Parent.Name, 'HDUNUM:', self.Parent.HduNum

        # Physical file pointer and HduNum:
        fptr, pHduNum = self.HduMap[self.Parent.HduNum]

        status = cfitsio.fits_movabs_hdu(fptr,pHduNum)

        # Create the set of functions needed to write one column
        # add the function needed for the type of the actual column
        if len(self._writeColFunctions) == 0:
            self._createWriteColFunctions()
        # Same for mapping from FITSIO Datatypes to Numeric Typecodes:
        if len(self._colNumericTypecodes) == 0:
            self._createTypecodes()

        # Get numbers of rows allready present and of number of columns
        # from the table:
        status, numRowsPresent = cfitsio.fits_get_num_rows(fptr)
        status, numColsTable = cfitsio.fits_get_num_cols(fptr)

        # Get number of rows and columns present in naData
        numRowsData, numColsData = naData.shape

        # Check that the numbers of columns in naData and the table
        # are the same:
        if numColsData != numColsTable:
            msg =  "Number of columns in Data (%d) " % numColsData
            msg += "does not match those in table (%d)" % numColsTable
            myError = MBFITSError(msg)
            raise myError

        startRow = numRowsPresent + 1

##         print "RS_DEBUG startRow, numRowsData, numColsData:", startRow, numRowsData, numColsData
                    
        for iCol0 in xrange(numColsData):
            iCol1 = iCol0 + 1

            # Cast the data in this column to the appropriate type:
            colData = naData[:,iCol0].astype(self._colNumericTypecodes[iCol0])

##             print "RS_DEBUG iCol0, self._colNumericTypecodes[iCol0], colData",  iCol0, self._colNumericTypecodes[iCol0], colData

            # Write this column:
            status = self._writeColFunctions[iCol0] \
                     (fptr, iCol1, startRow, 1, numRowsData, colData)

##             print "RS_DEBUG status:", status
            
        if config.DEBUG:
            print '.......... '+str(numRowsData)+' rows added to table '+self.Parent.Name+'. '

    def addTableRowsArraydata(self, naData):

        """
        Specialized method to fill the ARRAYDATA table.

        naData is a Numeric.array of shape (numRowsData, numColsData).

        The slice naData[:,0] contains the data to go into the first column
        of ARRAYDATA, the slice naData[:,1:] contains the data for the second
        column.

        Since at this time the methods BinTable._getTFORM, ._getTDIM, and
        ._getDataDimension, which could be used to check the correct
        shape of naData, are broken, this check canot be performed. 

        The slices of naData for individual columns will be casted to the
        datatype as defined in the XML file.
        """

        if not self.HduMap:
            msg = 'No HduMap. File seems to be closed.'
            myError = MBFITSError(msg)
            raise myError
        
## 	print 'DEBUG: ADDTABLEROWSARAYDATA NAME:', self.Parent.Name, 'HDUNUM:', self.Parent.HduNum

        # Physical file pointer and HduNum:
        fptr, pHduNum = self.HduMap[self.Parent.HduNum]

        status = cfitsio.fits_movabs_hdu(fptr,pHduNum)

        # Create the set of functions needed to write one column
        # add the function needed for the type of the actual column
        if len(self._writeColFunctions) == 0:
            self._createWriteColFunctions()
        # Same for mapping from FITSIO Datatypes to Numeric Typecodes:
        if len(self._colNumericTypecodes) == 0:
            self._createTypecodes()

        # Get numbers of rows allready present and of number of columns
        status, numRowsPresent = cfitsio.fits_get_num_rows(fptr)

        # Get number of rows and columns present in naData
        numRowsData, numColsData = naData.shape

        startRow = numRowsPresent + 1

##         print "RS_DEBUG startRow, numRowsData, numColsData:", startRow, numRowsData, numColsData
                    
        # Column 1:
        # Cast the data in this column to the appropriate type:
        colData = naData[:,0].astype(self._colNumericTypecodes[0])
        # Write this column:
        status = self._writeColFunctions[0] \
                 (fptr, 1, startRow, 1, numRowsData, colData)
##         print "RS_DEBUG Column 1 status:", status

        # Column 2:
        # Cast the data in this column to the appropriate type:
        colData = naData[:,1:].astype(self._colNumericTypecodes[1])
        # Write this column:
        status = self._writeColFunctions[0] \
                 (fptr, 2, startRow, 1, Numeric.size(colData), colData)
##         print "RS_DEBUG Column 2 status:", status
            
        if config.DEBUG:
            print '.......... '+str(numRowsData)+' rows added to table '+self.Parent.Name+'. '


    def updateTableColumn(self,colName,dataList,firstRow=None,lastRow=None):

        """ Add or update data in a table column """
        
        if not self.HduMap:
            msg = 'No HduMap. File seems to be closed.'
            myError = MBFITSError(msg)
            raise myError

        # First set the functions for column writing, so no if clause is needed
        # when caling this method multiple times

        # Physical file pointer and HduNum:
        fptr, pHduNum = self.HduMap[self.Parent.HduNum]

        status = cfitsio.fits_movabs_hdu(fptr,pHduNum)
        if len(self._writeColFunctions) == 0:
            self._createWriteColFunctions()
        
        status, numRows = cfitsio.fits_get_num_rows(fptr)
        if firstRow == None:
            firstRow = numRows+1
        numElements = len(dataList)  # number of input elements
        if lastRow:
            if (lastRow-firstRow+1) != numElements:
                msg = 'Number of rows to add does not fit to length of dataList.'
                myError = MBFITSError(msg)
                raise myError
        else:
            lastRow = firstRow+numElements

        status, colNum = cfitsio.fits_get_colnum(fptr,1,colName)
        if not status:
            # WARNING: needs a test for correct datatype in dataList to fit to column type
            # - may crash here till this is done
            status = self._writeColFunctions[colNum-1](fptr,colNum,firstRow,1,numElements,dataList)
            if status:
                msg = "Wasn't able to update column %s (CFitsio status code %s)"%(colName,str(status))
                myError = MBFITSError(msg)
                raise myError
        else:
            msg = 'Column (%s) not found.'% colName
            myError = MBFITSError(msg)
            raise myError
        
        return lastRow        



    def read(self,firstRow=1,lastRow=None):

        """ Read binary table columns from MBFits-File """

        if not self.HduMap:
            msg = 'No HduMap. File seems to be closed.'
            myError = MBFITSError(msg)
            raise myError
        
        # Physical file pointer and HduNum:
        fptr, pHduNum = self.HduMap[self.Parent.HduNum]

        status,hdu = cfitsio.fits_movabs_hdu(fptr,pHduNum)
        
	# Get the number of columns (= fields) and the number of rows
	status, numCol = cfitsio.fits_get_num_cols(fptr)
	status, numRow = cfitsio.fits_get_num_rows(fptr)
        if lastRow:
            if lastRow >numRow:
                msg = 'Last row larger than total rows available.'
                myError = MBFITSError(msg)
                raise myError
            else:
                numRow = lastRow
            
        if firstRow != 1:
            if firstRow <1:
                msg = 'First row < 1.'
                myError = MBFITSError(msg)
                raise myError
            elif firstRow > numRow:
                msg = 'First row larger than last row.'
                myError = MBFITSError(msg)
                raise myError
        
        # Append the whole content of each column, one after the other,
	# in the data attribute of the BinTable object
        dataSet = {}
	for i in xrange(numCol):
            if config.DEBUG > 1:
	        print "Reading column number ",i+1," TTYPE: ",self._getTTYPE(i+1),
                print "TFORM ",self._getTFORM(i+1),
                print "TDIM ",self._getTDIM(i+1)
            
            data = self.readTableColumn(self._getTTYPE(i+1),numRow,firstRow)
            dataSet[self.ttypes[i]] = data
            
        if firstRow and lastRow:
            return dataSet
        else:
            self.Data = dataSet



    def _colData2rowData(self,dataCols):

        """ Transposes a column organized data dictionary into a list of rows and returns it."""

        if not self.HduMap:
            msg = 'No HduMap. File seems to be closed.'
            myError = MBFITSError(msg)
            raise myError
        
        # Physical file pointer and HduNum:
        fptr, pHduNum = self.HduMap[self.Parent.HduNum]
        
        dataRows = []
        status, numCols = cfitsio.fits_get_num_cols(fptr)            
        for rowNum in xrange(len(dataCols.values()[0])):
            rowData = []
            for colNum in xrange(1,numCols+1):
                colName = self._getTTYPE(colNum)
                col = dataCols[colName]
                rowData.append(col[rowNum])
            dataRows.append(rowData)
            
        return dataRows



    def readTableColumn(self,colName,numRow=0,firstRow=1):

        """ Read content of column 'colName' from a Binary Table.

        Input arguments:
        colName:  name of the column to be read
        numRow:   number of rows to be read
        firstRow: first row to be read

        returns a list containing the values of that column from firstRow to firstRow+numRow
        """
        
        
        if not self.HduMap:
            msg = 'No HduMap. File seems to be closed.'
            myError = MBFITSError(msg)
            raise myError
        
        # Physical file pointer and HduNum:
        fptr, pHduNum = self.HduMap[self.Parent.HduNum]
        
        status,hdu = cfitsio.fits_movabs_hdu(fptr,pHduNum)

        if len(self._readColFunctions) == 0:
            self._createReadColFunctions()
        
        # if the number of rows is not input to the method, get it from fits_get_num_rows
        status, numTableRows = cfitsio.fits_get_num_rows(fptr)
        if numRow == 0:
            numRow = numTableRows

        # Test if first row number is valid
        if firstRow != 1:
            if firstRow <1:
                msg = 'First row number passed < 1. Must be 1 or higher.'
                myError = MBFITSError(msg)
                raise myError
            elif firstRow > numTableRows:
                msg = 'First row larger than last row number.'
                myError = MBFITSError(msg)
                raise myError
        # Test if last row number is valid
        if (firstRow + numRow -1) > (numTableRows):
            msg = 'Last row number to read larger than number of rows in table.'
            myError = MBFITSError(msg)
            raise myError
        
        if not numRow:
            msg = 'No data found in table '+self.Parent.Name+'.'
            print 'WARNING:',msg
            result = []

        elif numRow > 0:
            status, colNum = cfitsio.fits_get_colnum(fptr,0,colName)
            if status:
                msg = "Column "+colName+" not found"
                myError = MBFITSError(msg)
                raise myError

            # Handle non-scalar data: read the TFORMi keyword
            colType = self._getTFORM(colNum)

            # Then get the dimension from the coltype description
            dataDim = self._getDataDimension(colType)

            # Read row by row for strings and variable length arrays
            if ((colType[0]=='1' and colType[1] == 'P') or colType[-1] == 'A'):
                result=[]
               
                for rowNum in xrange(firstRow,firstRow+numRow):
                    # Use the appropriate fits_read_col method, according to data type
                    status, tmp = self._readColFunctions[colNum-1](fptr,colNum,rowNum,1,dataDim,0)
                    result.append(tmp)
            # all other types can be read all rows at once
            else:
                status, result = self._readColFunctions[colNum-1](fptr,colNum,firstRow,1,numRow*dataDim,0)            

            #if string.find(colType,'A') < 0:
            #    # cfitsio uses Numeric.
            #    result = Numeric.array([data])[0]
            #else:
            #    result = data

            # Now, if there is a TDIM field, the array should be reshaped

            # First try to get the dimension information from memory
            tdim = self._getTDIM(colNum)
                    
            if tdim:
                dim = tdim[1:len(tdim)-1]	# remove parenthesis
                # Let's check if the size of the data is consistent with this tdim
                # compute the expected size of data for one row, as represented by tdim:
                # lookfor commas (,) and compute the product of each comma separated value
                prod,curPos,nextPos = 1,0,string.find(dim,',')
                while nextPos >= 0:
                    number = string.atoi(dim[curPos:curPos+nextPos])
                    prod = prod*number
                    curPos += nextPos+1
                    nextPos = string.find(dim[curPos:len(dim)],',')
                # and the last dimension, after the last comma
                prod = prod*string.atoi(dim[curPos:len(dim)])

                # Now compute the size of what has been read
                readSize = Numeric.size(result)
                if readSize == prod*numRow:
                    cmd = 'shapeResult = Numeric.reshape(result,('
                    dimList=dim.split(',')
                  
                    dimList.reverse()
                    
                    dimListString=','.join(dimList)

                    cmd = cmd+str(numRow)+','+dimListString+'))'
                    exec(cmd)
                    result = shapeResult
                else:
                    print "Size of data (",readSize," values) not consistent with expected"
                    print "size (",prod," values) as given by the TDIM keyword:",tdim
                    print "Something might be wrong with the creation of the MB-FITS file!"
            elif (dataDim > 1 and string.find(colType,'A') < 0):
                cmd = 'shapeResult = Numeric.reshape(result,('
                cmd = cmd+str(numRow)+','+str(dataDim)+'))'
                exec(cmd)
                result = shapeResult
            # for variable length arrays that have dimension 1 this time
            elif (dataDim == 1 and string.find(colType,'A') < 0 and colType[0] == '1'):
                cmd = 'shapeResult = Numeric.reshape(result,('
                cmd = cmd+str(numRow)+','+str(dataDim)+'))'
                exec(cmd)
                result = shapeResult

        return result



    def readTableRow(self,numRow):

        """ Read one table row and returns a list of values. """

        if not self.HduMap:
            msg = 'No HduMap. File seems to be closed.'
            myError = MBFITSError(msg)
            raise myError
        
        # Physical file pointer and HduNum:
        fptr, pHduNum = self.HduMap[self.Parent.HduNum]
        
        # Move to the table's HDU
        status,hdu = cfitsio.fits_movabs_hdu(fptr,pHduNum)

        if not len(self._readColFunctions):
            self._createReadColFunctions()
        
        # get the number of rows in this table
	status, nbRow = cfitsio.fits_get_num_rows(fptr)

        # Check if the row requestet by numRow is in the range of available rows
	if (nbRow < numRow):
            msg = "Not enough rows in this table"
            myError = MBFITSError(msg)
            raise myError
        
        # Get the number of columns for this table
	status, numCol = cfitsio.fits_get_num_cols(fptr)
        
        #if config.DEBUG:
	#    print "BinTable.readTableRow: found numCol = ",numCol," column(s) to read"
	result = []
        
        # Loop over the number of columns to read the entries for each element of the row
	for colNum in xrange(1,numCol+1):
            #if config.DEBUG:
	    #    print "Reading column number ",colNum," ..."

            
            # Handle non-scalar data: read the TFORMi keyword
            colType = self._getTFORM(colNum)

            dataDim = self._getDataDimension(colType)

            # Use the appropriate fits_read_col method, according to data type
            status, tmp = self._readColFunctions[colNum-1](fptr,colNum,numRow,1,dataDim,0)
                    
            # But convert list to array if not a string and more than one value
	    if (Numeric.rank(tmp) and string.find(colType,'A') < 0):
		tmp = Numeric.array([tmp])[0]

            # Now, if there is a TDIM field, the array should be reshaped
            tdim = self._getTDIM(colNum)

	    if tdim:
	        dim = tdim[1:len(tdim)-1]	# remove parenthesis
                tmp = Numeric.reshape(tmp,(int(dim)))
                #cmd = 'shapeResult = Numeric.reshape(tmp,('+dim+'))'
                #exec(cmd)
                #tmp = shapeResult

	    result.append(tmp)

	return result



    def readTableRows(self,firstRow,lastRow):

        """ Read a set of table rows from firstRow to lastRow and returns a list of rows. """
        
        colData = self.read(firstRow,lastRow)
        rowData = self._colData2rowData(colData)
        return rowData



    def selectRows(self,expression,firstLine=1,lastLine=None):

        """ Return table rows according to expression (which has to be a string)\n
        The expression can be as complex as described in the cfitsio manual, as it is\n
        directly passed to the cfitsio routine that is being used by this method.\n
        Example: To select all rows from a MONITOR table whith the MONPOINT column\n
        matching ANTENNA_AZ_EL, you will have to pass the string 'MONPOINT=="ANTENNA_AZ_EL"'\n
        as the expression. See the cfitsio manual for further detail."""

        if not self.HduMap:
            msg = 'No HduMap. File seems to be closed.'
            myError = MBFITSError(msg)
            raise myError

        # Physical file pointer and HduNum:
        fptr, pHduNum = self.HduMap[self.Parent.HduNum]
        
        # Move to the table's HDUNum 
        status = cfitsio.fits_movabs_hdu(fptr,pHduNum)
        
        # Get total numer of rows
        status, totalRows = cfitsio.fits_get_num_rows(fptr)
        if lastLine:
            # Calculate number of rows to be evaluated
            numRows = lastLine - firstLine +1 # when both are equal, one line will be checked
        else:
            lastLine = totalRows
            numRows = totalRows - firstLine +1
            
        # Find rows according to expression
        status, numRowsFound, rowResults = cfitsio.fits_find_rows(fptr,expression,firstLine,numRows)

        result = []
        if not status:
            data = self.readTableRows(firstLine,lastLine)
            
            # As fits_find_rows returns a list with '1' or '0' (for true and false for each row)
            # we need to loop over the total rows
            i = firstLine
            # Loop over all rows
            for rowResult in rowResults:
                # If the row matches the expression
                if int(rowResult):
                    # get the row
                    row = data[i-1]
                    # and append the row + it's number (i) to the result
                    result.append(row+[i])
                i +=1
        else:            
            # Check for valid input of firstLine and lastLine and create an error message
            if firstLine > lastLine:
                msg = 'FirstLine larger than lastLine.'
            elif firstLine <1:
                msg = 'FirstLine must be an integer >0.'
            elif lastLine >totalRows:
                msg = 'Not enough rows (lastLine > total rows of this table).'
            # If the error can't be determined from the method's argument print the cfitsio status
            else:
                msg = "Unknown error: Cfitsio routine 'fits_find_rows' returned with status %s."%(str(status))
            myError = MBFITSError(msg)
            raise myError
            
        return result
        

        
#########################################################################
def _setType(Key):
#########################################################################
    
    """ Convert keyword to correct python datatype  """
    
    status, t = cfitsio.fits_get_keytype(Key)
    if t=='C':
        Result = Key[1:-1]
        Result = string.strip(Result)
    elif t == 'L':
        Result = Key
    elif t == 'I':
        Result = string.atoi(Key)
    elif t == 'F':
        Result = string.atof(Key)
    else:
        Result = Key
    
    return Result



#########################################################################
def _searchTable(Table,TableName):
#########################################################################

    """ recursive search for the table named TableName (string) from the table Table (Table object) """

    if (Table.Name == TableName):
        Result = Table
    else:
        Result = _searchTable(Table.Parent,TableName)
    return Result



#########################################################################
class MBFITSError(Exception):
#########################################################################
    
    def __init__(self, msg, code=None):
        self.code = code
        self.msg = msg

        
    def __str__(self):
        return `self.msg`



#########################################################################
def mergeMBFitsGroupToSingleFile(filenameIn, filenameOut):
#########################################################################

    """
    Copies the multifile structured MBFitsfile filenameIn to the singlefile
    structured MBFitsfile filenameOut.

    filenameIn may specify either the logical filename (directory only) of the
    input file or the physical filename (directoryname/GROUPING.fits).
    """

    # Allow directory name as filenameIn:
    if os.path.isdir(filenameIn):
        filenameIn = filenameIn + "/GROUPING.fits"

    # Open input file, create output file:
    status, fptrIn = cfitsio.fits_open_file(filenameIn, 0) # read-only
    if status:
        msg = "Error %d while opening file %s" % (status, filenameIn)
        raise msg

    status, fptrOut = cfitsio.fits_create_file(filenameOut)
    if status:
        msg = "Error %s while creating file %s" % (status, filenameOut)
        raise msg

    # Provide structural information about grouping tables:
    hg = _HierGroup()

    # Walk through input file:
    status, pHduCountIn = cfitsio.fits_get_num_hdus(fptrIn)
    if status:
        msg = "Error %d while reading file %s" % (status, filenameIn)
        raise msg

    for pHduNumIn in xrange(1, pHduCountIn+1):
        
        print "HDU %d/%d" % (pHduNumIn, pHduCountIn)
        
        status, cpHduNumIn = cfitsio.fits_movabs_hdu(fptrIn, pHduNumIn)
        if status:
            msg = "Error %d while moving in file %s to HDU %d" \
                  % (status, filenameIn, pHduNumIn)
            raise msg

        if hg.isGroupingTable(fptrIn):
            # Copy all members:
            status, nGroupMembers = cfitsio.fits_get_num_members(fptrIn)
            if status:
                msg = "Error %d while getting number of members" % status + \
                      "in HDU %d in file %s" % (pHduNumIn, filenameIn)
                raise msg

            for iGroupMember in xrange(1, nGroupMembers+1):

                print "  Member %d/%d" % (iGroupMember, nGroupMembers)  
                
                #define OPT_MCP_ADD     0
                #define OPT_MCP_NADD    1
                #define OPT_MCP_REPL    2
                #define OPT_MCP_MOV     3
                status = cfitsio.fits_copy_member(fptrIn, \
                                                  fptrOut, \
                                                  iGroupMember, \
                                                  3) # do not add new entry
                                                     # to grouping table 
                if status:
                    msg = "Error %d while copiing member %d in HDU %d" \
                          % (status, iGroupMember, pHduNumIn)
            
        else:
            # Copy HDU:
            status = cfitsio.fits_copy_hdu(fptrIn, fptrOut, 0) # no more keys
            if status:
                msg = "Error %d while copiing HDU %d" \
                      % (status, pHduNumIn)
                raise msg

    # Close files:
    status = cfitsio.fits_close_file(fptrOut)
    if status:
        msg = "Error %d while closing file %s" % (status, filenameOut)
        raise msg
    
    status = cfitsio.fits_close_file(fptrIn)
    if status:
        msg = "Error %d while closing file %s" % (status, filenameIn)
        raise msg
    
